#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os, sys, re
from prjxray import util

TAG_PART_RE = re.compile(r"^[a-zA-Z][0-9a-zA-Z_]*(\[[0-9]+\])?$")


def check_tag_name(tag):
    '''
    Checks if the tag name given by the used conforms to the valid fasm
    name rules.

    >>> check_tag_name("CELL.feature19.ENABLED")
    True
    >>> check_tag_name("FEATURE")
    True
    >>> check_tag_name("TAG.")
    False
    >>> check_tag_name(".TAG")
    False
    >>> check_tag_name("CELL..FEATURE")
    False
    >>> check_tag_name("CELL.3ENABLE")
    False
    >>> check_tag_name("FEATURE.12.ON")
    False
    '''

    for part in tag.split("."):
        if not len(part) or TAG_PART_RE.match(part) is None:
            return False

    return True


def run(fn_ins, fn_out, strict=False, track_origin=False, verbose=False):
    # tag to bits
    entries = {}
    # tag to (bits, line)
    tags = dict()
    # bits to (tag, line)
    bitss = dict()

    for fn_in in fn_ins:
        for line, (tag, bits, mode, origin) in util.parse_db_lines(fn_in):
            line = line.strip()
            assert mode is not None or mode != "always", "strict: got ill defined line: %s" % (
                line, )

            if not check_tag_name(tag):
                assert not strict, "strict: Invalid tag name '{}'".format(tag)

            if tag in tags:
                orig_bits, orig_line, orig_origin = tags[tag]
                if orig_bits != bits:
                    print(
                        "WARNING: got duplicate tag %s" % (tag, ),
                        file=sys.stderr)
                    print("  Orig line: %s" % orig_line, file=sys.stderr)
                    print("  New line : %s" % line, file=sys.stderr)
                    assert not strict, "strict: got duplicate tag"
                origin = os.path.basename(os.getcwd())
                if track_origin and orig_origin != origin:
                    origin = orig_origin + "," + origin
            if bits in bitss:
                orig_tag, orig_line = bitss[bits]
                if orig_tag != tag:
                    print(
                        "WARNING: got duplicate bits %s" % (bits, ),
                        file=sys.stderr)
                    print("  Orig line: %s" % orig_line, file=sys.stderr)
                    print("  New line : %s" % line, file=sys.stderr)
                    assert not strict, "strict: got duplicate bits"

            if track_origin and origin is None:
                origin = os.path.basename(os.getcwd())
            entries[tag] = (bits, origin)
            tags[tag] = (bits, line, origin)
            if bits != None:
                bitss[bits] = (tag, line)

    util.write_db_lines(fn_out, entries, track_origin)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Combine multiple .db files")

    util.db_root_arg(parser)
    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument('--track_origin', action='store_true', help='')
    parser.add_argument('--out', help='')
    parser.add_argument('ins', nargs='+', help='Last takes precedence')
    args = parser.parse_args()

    run(
        args.ins,
        args.out,
        strict=int(os.getenv("MERGEDB_STRICT", "1")),
        track_origin=args.track_origin,
        verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse
from prjxray.db import Database
from prjxray.grid import BlockType
from prjxray import util


def main():
    parser = argparse.ArgumentParser(
        description="Tool for checking which tiles have bits defined.")

    util.db_root_arg(parser)
    util.part_arg(parser)
    parser.add_argument('--show-only-missing', action='store_true')
    parser.add_argument('--verbose', action='store_true')

    args = parser.parse_args()

    db = Database(args.db_root, args.part)
    grid = db.grid()

    tile_types = {}
    for tile in grid.tiles():
        gridinfo = grid.gridinfo_at_tilename(tile)
        if gridinfo.tile_type not in tile_types:
            tile_types[gridinfo.tile_type] = []

        tile_types[gridinfo.tile_type].append((tile, gridinfo))

    total_tile_count = 0
    total_have_bits = 0

    for tile_type, tiles in sorted(tile_types.items()):
        try:
            tile_type_info = db.get_tile_type(tile_type)

            # Skip empty tiles, as no base address is requied.
            if len(tile_type_info.get_pips()) == 0 and len(
                    tile_type_info.get_sites()) == 0:
                continue
        except KeyError:
            pass

        # INT_INTERFACE tiles likely don't contain configuration?  Remove this
        # if this ends up false.
        if 'INT_INTERFACE' in tile_type:
            continue

        if 'BRKH' in tile_type:
            continue

        have_bits = 0
        for tile_name, gridinfo in tiles:
            total_tile_count += 1
            if BlockType.CLB_IO_CLK in gridinfo.bits:
                have_bits += 1
                total_have_bits += 1

        if args.show_only_missing and have_bits == len(tiles):
            continue

        print(
            '{}: {}/{} ({:.2f} %)'.format(
                tile_type, have_bits, len(tiles),
                100. * float(have_bits) / len(tiles)))

        if args.verbose:
            tiles_with_missing_bits = []
            for tile_name, gridinfo in tiles:
                total_tile_count += 1
                if BlockType.CLB_IO_CLK not in gridinfo.bits:
                    tiles_with_missing_bits.append(tile_name)

            for tile_name in sorted(tiles_with_missing_bits):
                print('{} is missing CLB_IO_CLK'.format(tile_name))

    print('')
    print(
        'Summary: {}/{} ({:.2f} %)'.format(
            total_have_bits, total_tile_count,
            100. * float(total_have_bits) / total_tile_count))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os, random
random.seed(0)
from prjxray import util
from prjxray import verilog

# INCREMENT is the amount of additional CLBN to be instantiated in the design.
# This makes the fuzzer compilation more robust against failures.
INCREMENT = os.getenv('CLBN', 0)
CLBN = 600 + int(INCREMENT)
print('//Requested CLBs: %s' % str(CLBN))


def gen_slicels():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites(
        ['SLICEL']):
        yield site_name


def gen_slicems():
    for _tile_name, site_name, _site_type in util.get_roi().gen_sites(
        ['SLICEM']):
        yield site_name


DIN_N = CLBN * 8
DOUT_N = CLBN * 8

verilog.top_harness(DIN_N, DOUT_N)

f = open('params.csv', 'w')
f.write('module,loc,n\n')
slicels = gen_slicels()
slicems = gen_slicems()
print(
    'module roi(input clk, input [%d:0] din, output [%d:0] dout);' %
    (DIN_N - 1, DOUT_N - 1))
for i in range(CLBN):

    use_slicem = (i % 2) == 0

    if use_slicem:
        loc = next(slicems)
        variants = ['CY', 'F78', 'O5', 'XOR', 'B5Q', 'MC31']
    else:
        loc = next(slicels)
        variants = ['CY', 'F78', 'O5', 'XOR', 'B5Q']

    # Don't have an O6 example
    modules = ['clb_NOUTMUX_' + x for x in variants]
    module = random.choice(modules)

    if module == 'clb_NOUTMUX_MC31':
        n = 3  # Only DOUTMUX has MC31 input
    elif module == 'clb_NOUTMUX_F78':
        n = random.randint(0, 2)
    else:
        n = random.randint(0, 3)

    print('    %s' % module)
    print('            #(.LOC("%s"), .N(%d))' % (loc, n))
    print(
        '            clb_%d (.clk(clk), .din(din[  %d +: 8]), .dout(dout[  %d +: 8]));'
        % (i, 8 * i, 8 * i))

    f.write('%s,%s,%s\n' % (module, loc, n))
f.close()
print(
    '''endmodule

// ---------------------------------------------------------------------

''')

print(
    '''
module myLUT8 (input clk, input [7:0] din,
        output lut8o, output lut7bo, output lut7ao,
        //caro: XOR additional result (main output)
        //carco: CLA result (carry module additional output)
        output caro, output carco,
        output bo5, output bo6,
        //Note: b5ff_q requires the mux and will conflict with other wires
        //Otherwise this FF drops out
        //output wire [3:0] n5ff_q);
        output wire ff_q,
        output wire mc31);

    parameter N=-1;
    parameter LOC="SLICE_FIXME";
    parameter ALUT_SRL=0;

    wire [3:0] caro_all;
    assign caro = caro_all[N];
    wire [3:0] carco_all;
    assign carco = carco_all[N];

    wire [3:0] lutno6;
    assign bo6 = lutno6[N];
    wire [3:0] lutno5;
    assign bo5 = lutno5[N];

    //Outputs does not have to be used, will stay without it
    (* LOC=LOC, BEL="F8MUX", KEEP, DONT_TOUCH *)
    MUXF8 mux8 (.O(lut8o), .I0(lut7bo), .I1(lut7ao), .S(din[6]));
    (* LOC=LOC, BEL="F7BMUX", KEEP, DONT_TOUCH *)
    MUXF7 mux7b (.O(lut7bo), .I0(lutno6[3]), .I1(lutno6[2]), .S(din[6]));
    (* LOC=LOC, BEL="F7AMUX", KEEP, DONT_TOUCH *)
    MUXF7 mux7a (.O(lut7ao), .I0(lutno6[1]), .I1(lutno6[0]), .S(din[6]));

	(* LOC=LOC, BEL="D6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_DEAD_0000_0001)
	) lutd (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutno5[3]),
		.O6(lutno6[3]));

	(* LOC=LOC, BEL="C6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_BEEF_0000_0001)
	) lutc (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutno5[2]),
		.O6(lutno6[2]));

	(* LOC=LOC, BEL="B6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_CAFE_0000_0001)
	) lutb (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutno5[1]),
		.O6(lutno6[1]));

    generate if (ALUT_SRL != 0) begin

    (* LOC=LOC, BEL="A6LUT", KEEP, DONT_TOUCH *)
	SRLC32E #(
		.INIT(64'h8000_1CE0_0000_0001)
	) srla (
        .CLK(clk),
        .CE(din[6]),
        .D(din[5]),
		.A(din[4:0]),
		.Q(lutno6[0]),
        .Q31(mc31));

    assign lutno5[0] = din[6];

    end else begin

    (* LOC=LOC, BEL="A6LUT", KEEP, DONT_TOUCH *)
	LUT6_2 #(
		.INIT(64'h8000_1CE0_0000_0001)
	) luta (
		.I0(din[0]),
		.I1(din[1]),
		.I2(din[2]),
		.I3(din[3]),
		.I4(din[4]),
		.I5(din[5]),
		.O5(lutno5[0]),
		.O6(lutno6[0]));

    end endgenerate

    //Outputs do not have to be used, will stay without them
	(* LOC=LOC, KEEP, DONT_TOUCH *)
    CARRY4 carry4(.O(caro_all), .CO(carco_all), .DI(lutno5), .S(lutno6), .CYINIT(1'b0), .CI());

    generate
        if (N == 3) begin
	        (* LOC=LOC, BEL="D5FF", KEEP, DONT_TOUCH *)
	        FDPE d5ff (
		        .C(clk),
		        .Q(ff_q),
		        .CE(1'b1),
		        .PRE(1'b0),
		        .D(lutno5[3]));
        end
        if (N == 2) begin
	        (* LOC=LOC, BEL="C5FF", KEEP, DONT_TOUCH *)
	        FDPE c5ff (
		        .C(clk),
		        .Q(ff_q),
		        .CE(1'b1),
		        .PRE(1'b0),
		        .D(lutno5[2]));
        end
        if (N == 1) begin
	        (* LOC=LOC, BEL="B5FF", KEEP, DONT_TOUCH *)
	        FDPE b5ff (
		        .C(clk),
		        .Q(ff_q),
		        .CE(1'b1),
		        .PRE(1'b0),
		        .D(lutno5[1]));
        end
        if (N == 0) begin
	        (* LOC=LOC, BEL="A5FF", KEEP, DONT_TOUCH *)
	        FDPE a5ff (
		        .C(clk),
		        .Q(ff_q),
		        .CE(1'b1),
		        .PRE(1'b0),
		        .D(lutno5[0]));
        end
    endgenerate
endmodule

//******************************************************************************
//BOUTMUX tests

module clb_NOUTMUX_CY (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=1;

    myLUT8 #(.LOC(LOC), .N(N))
            myLUT8(.clk(clk), .din(din), .lut8o(),
            .caro(), .carco(dout[0]),
            .bo5(), .bo6(),
            .ff_q());
endmodule

//clb_NOUTMUX_F78: already have above as clb_LUT8
module clb_NOUTMUX_F78 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=1;
    wire lut8o, lut7bo, lut7ao;
    /*
    D: N/A (no such mux position)
    C: F7B:O
    B: F8:O
    A: F7A:O
    */
    generate
        if (N == 3) begin
            //No muxes, so this is undefined
           invalid_configuration invalid_configuration3();
        end else if (N == 2) begin
            assign dout[0] = lut7bo;
        end else if (N == 1) begin
            assign dout[0] = lut8o;
        end else if (N == 0) begin
            assign dout[0] = lut7ao;
        end
    endgenerate

    myLUT8 #(.LOC(LOC), .N(N))
            myLUT8(.clk(clk), .din(din),
            .lut8o(lut8o), .lut7bo(lut7bo), .lut7ao(lut7ao),
            .caro(), .carco(),
            .bo5(), .bo6(),
            .ff_q());
endmodule

module clb_NOUTMUX_O5 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=1;

    myLUT8 #(.LOC(LOC), .N(N))
            myLUT8(.clk(clk), .din(din), .lut8o(),
            .caro(), .carco(),
            .bo5(dout[0]), .bo6(),
            .ff_q());
endmodule

/*
//FIXME: need to force it to use both X and O6
module clb_NOUTMUX_O6 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=1;

    myLUT8 #(.LOC(LOC), .N(N))
            myLUT8(.clk(clk), .din(din), .lut8o(), .co(), .carco(), .bo5(), .bo6());
endmodule
*/

module clb_NOUTMUX_XOR (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=1;

    myLUT8 #(.LOC(LOC), .N(N))
            myLUT8(.clk(clk), .din(din), .lut8o(),
            .caro(dout[0]), .carco(),
            .bo5(), .bo6(),
            .ff_q());
endmodule

module clb_NOUTMUX_B5Q (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=1;

    myLUT8 #(.LOC(LOC), .N(N))
            myLUT8(.clk(clk), .din(din),
            .lut8o(),
            .caro(), .carco(),
            .bo5(), .bo6(),
            .ff_q(dout[0]));
endmodule

module clb_NOUTMUX_MC31 (input clk, input [7:0] din, output [7:0] dout);
    parameter LOC="SLICE_FIXME";
    parameter N=0; // Dummy

    myLUT8 #(.LOC(LOC), .N(0), .ALUT_SRL(1))
            myLUT8(.clk(clk), .din(din),
            .lut8o(),
            .caro(), .carco(),
            .bo5(), .bo6(),
            .ff_q(), .mc31(dout[0]));
endmodule
''')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Generates a missing feature/bit report for LiteX design.

This script is fairly fragile, because it depends on the specific observation
that all of the remaining bits appear to either belong to HCLK_IOI or IOI3
tiles.  A more general version of this script could be created, but that was
not the point of this script.

"""
from fasm import parse_fasm_filename


def main():
    fasm_file = 'top.fasm'
    fasm_model = list(parse_fasm_filename(fasm_file))

    unknown_bits = {
        'HCLK_IOI': {},
        'IOI3': {},
    }

    total_unknown = 0
    for l in fasm_model:
        if l.annotations is None:
            continue

        annotations = {}
        for annotation in l.annotations:
            annotations[annotation.name] = annotation.value

        if 'unknown_bit' not in annotations:
            continue

        total_unknown += 1

        frame, word, bit = annotations['unknown_bit'].split('_')

        frame = int(frame, 16)
        word = int(word)
        bit = int(bit)

        frame_offset = frame % 0x80
        base_frame = frame - frame_offset

        # All remaining LiteX bits appear to be in this one IO bank, so limit
        # the tool this this one IO bank.
        assert base_frame == 0x00401580, hex(frame)

        SIZE = 4
        INITIAL_OFFSET = -2

        if word == 50:
            group = 'HCLK_IOI'
            offset = 45
        elif word < 50:
            group = 'IOI3'
            offset = ((word - INITIAL_OFFSET) // SIZE) * SIZE + INITIAL_OFFSET
        else:
            group = 'IOI3'
            word -= 1
            offset = ((word - INITIAL_OFFSET) // SIZE) * SIZE + INITIAL_OFFSET
            offset += 1
            word += 1

        bit = '{}_{:02d}'.format(
            frame_offset,
            (word - offset) * 32 + bit,
        )

        if bit not in unknown_bits[group]:
            unknown_bits[group][bit] = 0
        unknown_bits[group][bit] += 1

    print('Total unknown bits: {}'.format(total_unknown))
    for group in unknown_bits:
        print('Group {} (count = {}):'.format(group, len(unknown_bits[group])))
        for bit in sorted(unknown_bits[group]):
            print('  {} (count = {})'.format(bit, unknown_bits[group][bit]))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys, re
from prjxray.util import OpenSafeFile, db_root_arg, parse_db_line


def run(fnin, fnout=None, strict=False, verbose=False):
    with OpenSafeFile(fnin) as f:
        lines = f.read().split('\n')
    tags = dict()
    bitss = dict()
    for line in lines:
        line = line.strip()
        if line == '':
            continue
        # TODO: figure out what to do with masks
        if line.startswith("bit "):
            continue
        tag, bits, mode, _ = parse_db_line(line)
        if strict:
            if mode != "always":
                assert not mode, "strict: got ill defined line: %s" % (line, )
            if tag in tags:
                print("Original line: %s" % tags[tag], file=sys.stderr)
                print("New line: %s" % line, file=sys.stderr)
                assert 0, "strict: got duplicate tag %s" % (tag, )
            assert bits not in bitss, "strict: got duplicate bits %s: %s %s" % (
                bits, tag, bitss[bits])
        tags[tag] = line
        if bits != None:
            bitss[bits] = tag

    if fnout:
        with OpenSafeFile(fnout, "w") as fout:
            for line in sorted(lines):
                line = line.strip()
                if line == '':
                    continue
                fout.write(line + '\n')


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Parse a db file, checking for consistency")

    db_root_arg(parser)
    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument(
        '--strict',
        action='store_true',
        help='Complain on unresolved entries (ex: <0 candidates>, <const0>)')
    parser.add_argument('fin', help='')
    parser.add_argument('fout', nargs='?', help='')
    args = parser.parse_args()

    run(args.fin, args.fout, strict=args.strict, verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
This script allows to load a number of .db or .rdb files and display bits in
a nice visualization.

When more than one files are loaded, a difference between them is shown.
Differring bits are highlighted.
'''
import argparse
import re
import sys

import itertools

from prjxray.util import OpenSafeFile

# =============================================================================


def tagmap(tag):
    """
    Maps a specific tag name to its generic name
    """

    tag = tag.replace("CLBLL_L", "CLB")
    tag = tag.replace("CLBLL_M", "CLB")
    tag = tag.replace("CLBLM_L", "CLB")
    tag = tag.replace("CLBLM_M", "CLB")

    tag = tag.replace("SLICEL", "SLICE")
    tag = tag.replace("SLICEM", "SLICE")

    tag = tag.replace("LIOB33", "IOB33")
    tag = tag.replace("RIOB33", "IOB33")

    tag = tag.replace("LIOI3", "IOI3")
    tag = tag.replace("RIOI3", "IOI3")

    # TODO: Add other tag mappings

    return tag


def parse_bit(bit):
    """
    Decodes string describing a bit. Returns a tuple (frame, bit, value)
    """
    match = re.match("^(!?)([0-9]+)_([0-9]+)$", bit)
    assert match != None, bit

    val = int(match.group(1) != "!")
    frm = int(match.group(2))
    bit = int(match.group(3))

    return frm, bit, val


def load_and_sort_segbits(file_name, tagmap=lambda tag: tag):
    """
    Loads a segbit file (.db or .rdb). Skips bits containing '<' or '>'
    """

    # Load segbits
    segbits = {}
    with OpenSafeFile(file_name, "r") as fp:
        lines = fp.readlines()

        # Parse lines
        for line in lines:
            line = line.strip()
            fields = line.split()

            if len(fields) < 2:
                print("Malformed line: '%s'" % line)
                continue

            # Map name
            feature = tagmap(fields[0])

            # Decode bits
            bits = []
            for bit in fields[1:]:
                if "<" in bit or ">" in bit:
                    continue
                bits.append(parse_bit(bit))

            # Sort bits
            bits.sort(key=lambda bit: (bit[0], bit[1],))
            segbits[feature] = bits

    return segbits


# =============================================================================


def make_header_lines(all_bits):
    """
    Formats header lines
    """
    lines = []

    # Bit names
    bit_names = ["%d_%d" % (b[0], b[1]) for b in all_bits]
    bit_len = 6
    for i in range(bit_len):
        line = ""
        for j in range(len(all_bits)):
            bstr = bit_names[j].ljust(bit_len).replace("_", "|")
            line += bstr[i]
        lines.append(line)

    return lines


def make_data_lines(all_tags, all_bits, segbits):
    """
    Formats data lines
    """
    lines = []

    def map_f(b):
        if b < 0: return "0"
        if b > 0: return "1"
        return "-"

    # Bit data
    for tag in all_tags:
        if tag in segbits.keys():
            lines.append("".join(map(map_f, segbits[tag])))
        else:
            lines.append(" " * len(all_bits))

    return lines


def main():

    # Colors for TTY
    if sys.stdout.isatty():
        colors = {
            "NONE": "\033[0m",
            "DIFF": "\033[1m",
        }

    # Colors for pipe
    else:
        colors = {
            "NONE": "",
            "DIFF": "",
        }

    # ........................................................................

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("files", nargs="*", type=str, help="Input files")

    args = parser.parse_args()

    # Load segbits
    all_segbits = []
    for f in args.files:
        all_segbits.append(load_and_sort_segbits(f, tagmap))

    # List of all features and all bits
    all_tags = set()
    all_bits = set()

    for segbits in all_segbits:
        all_tags |= set(segbits.keys())
        for bits in segbits.values():
            all_bits |= set([(b[0], b[1]) for b in bits])

    all_tags = sorted(list(all_tags))
    all_bits = sorted(list(all_bits))

    # Convert bit lists to bit vectors
    for segbits in all_segbits:
        for tag in segbits.keys():
            vec = list([0] * len(all_bits))
            for i, bit in enumerate(all_bits):
                if (bit[0], bit[1], 0) in segbits[tag]:
                    vec[i] = -1
                if (bit[0], bit[1], 1) in segbits[tag]:
                    vec[i] = +1
            segbits[tag] = vec

    # Make header and data lines
    header_lines = make_header_lines(all_bits)
    data_lines = [
        make_data_lines(all_tags, all_bits, segbits) for segbits in all_segbits
    ]

    # Display
    max_tag_len = max([len(tag) for tag in all_tags])

    for l in header_lines:
        line = " " * max_tag_len + " "
        for i in range(len(all_segbits)):
            line += l + " "
        print(line)

    data_len = len(all_bits)
    for i, tag in enumerate(all_tags):
        line = tag.ljust(max_tag_len) + " "

        diff = bytearray(data_len)
        for l1, l2 in itertools.combinations(data_lines, 2):
            for j in range(data_len):
                if l1[i][j] != l2[i][j]:
                    diff[j] = 1

        for l in data_lines:
            for j in range(data_len):
                if diff[j]:
                    line += colors["DIFF"] + l[i][j] + colors["NONE"]
                else:
                    line += l[i][j]
            line += " "

        print(line)


# =============================================================================

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import io
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import lut_maker
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    '''
    IOB33S: main IOB of a diff pair
    IOB33M: secondary IOB of a diff pair
    IOB33: not a diff pair. Relatively rare (at least in ROI...2 of them?)
    Focus on IOB33S to start
    '''
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        sites = {}
        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['IOB33S', 'IOB33M']:
                sites[site_type] = site_name

        if sites:
            yield tile_name, sites


def write_params(params):
    pinstr = 'tile,site,pin,iostandard,drive,slew\n'
    for vals in params:
        pinstr += ','.join(map(str, vals)) + '\n'

    open('params.csv', 'w').write(pinstr)


def run():
    tile_types = ['IBUF', 'OBUF', 'IOBUF_INTERMDISABLE', None, None]

    i_idx = 0
    o_idx = 0
    io_idx = 0

    iostandards = [
        'LVCMOS12',
        'LVCMOS15',
        'LVCMOS18',
        'LVCMOS25',
        'LVCMOS33',
        'LVTTL',
        'SSTL135',
        'SSTL15',
        'LVDS_25',
        'TMDS_33',
    ]

    diff_map = {
        "SSTL135": ["DIFF_SSTL135"],
        "SSTL15": ["DIFF_SSTL15"],
    }

    only_diff_map = {
        "LVDS_25": ["LVDS_25"],
        "TMDS_33": ["TMDS_33"],
    }

    IN_TERM_ALLOWED = [
        'SSTL15',
        'SSTL15_R',
        'SSTL18',
        'SSTL18_R',
        'SSTL135',
        'SSTL135_R',
        'HSTL_I'
        'HSTL_I_18'
        'HSTL_II',
        'HSTL_II_18',
    ]

    slews = ['FAST', 'SLOW']
    pulls = ["NONE", "KEEPER", "PULLDOWN", "PULLUP"]

    luts = lut_maker.LutMaker()

    connects = io.StringIO()

    tile_params = []
    params = {
        "tiles": [],
        'INTERNAL_VREF': {},
    }

    with open(os.path.join(os.getenv('FUZDIR'), 'build', 'iobanks.txt')) as f:
        iobanks = set()
        iob_sites = dict()
        for l in f:
            fields = l.split(',')
            iob_site = fields[0]
            iobank = fields[1].rstrip()

            iobanks.add(iobank)
            iob_sites[iob_site] = iobank

    params['iobanks'] = iobanks

    iostandard_map = dict()
    for iobank in iobanks:
        iostandard = random.choice(iostandards)
        if iostandard in ['SSTL135', 'SSTL15']:
            params['INTERNAL_VREF'][iobank] = random.choice(
                (
                    .600,
                    .675,
                    .75,
                    .90,
                ))

        iostandard_map[iobank] = iostandard

    params['iobanks'] = list(iobanks)

    any_idelay = False
    for tile, sites in gen_sites():
        iostandard = None

        site_bels = {}
        for site_type, site_name in sites.items():
            iobank = iob_sites[site_name]
            iostandard = iostandard_map[iobank]

            if iostandard in ['LVTTL', 'LVCMOS18']:
                drives = [4, 8, 12, 16, 24]
            elif iostandard in ['LVCMOS12']:
                drives = [4, 8, 12]
            elif iostandard in ['SSTL135', 'SSTL15', 'LVDS_25', 'TMDS_33']:
                drives = None
            else:
                drives = [4, 8, 12, 16]

            if site_type.endswith('M'):
                if iostandard in diff_map:
                    site_bels[site_type] = random.choice(
                        tile_types + ['IBUFDS', 'OBUFDS', 'OBUFTDS'])
                elif iostandard in only_diff_map:
                    site_bels[site_type] = random.choice(
                        ['IBUFDS', 'OBUFDS', 'OBUFTDS', None, None])
                else:
                    site_bels[site_type] = random.choice(tile_types)
                is_m_diff = site_bels[site_type] is not None and site_bels[
                    site_type].endswith('DS')
            else:
                site_bels[site_type] = random.choice(tile_types)

        if is_m_diff or iostandard in only_diff_map:
            site_bels['IOB33S'] = None

        for site_type, site in sites.items():
            p = {}
            p['tile'] = tile
            p['site'] = site
            p['type'] = site_bels[site_type]

            if p['type'] is not None and p['type'].endswith('DS'):
                if iostandard in diff_map:
                    iostandard_site = random.choice(diff_map[iostandard])
                elif iostandard in only_diff_map:
                    iostandard_site = random.choice(only_diff_map[iostandard])
                p['pair_site'] = sites['IOB33S']
            else:
                iostandard_site = iostandard

            p['IOSTANDARD'] = verilog.quote(iostandard_site)
            p['PULLTYPE'] = verilog.quote(random.choice(pulls))

            if p['type'] is None:
                p['pad_wire'] = None
            elif p['type'] == 'IBUF':
                p['pad_wire'] = 'di[{}]'.format(i_idx)
                p['IDELAY_ONLY'] = random.randint(0, 1)
                if not p['IDELAY_ONLY']:
                    p['owire'] = luts.get_next_input_net()
                else:
                    any_idelay = True
                    p['owire'] = 'idelay_{site}'.format(**p)

                p['DRIVE'] = None
                p['SLEW'] = None
                p['IBUF_LOW_PWR'] = random.randint(0, 1)

                if iostandard in IN_TERM_ALLOWED:
                    p['IN_TERM'] = random.choice(
                        (
                            'NONE',
                            'UNTUNED_SPLIT_40',
                            'UNTUNED_SPLIT_50',
                            'UNTUNED_SPLIT_60',
                        ))

                i_idx += 1
            elif p['type'] == 'IBUFDS':
                p['pad_wire'] = 'di[{}]'.format(i_idx)
                i_idx += 1
                p['bpad_wire'] = 'di[{}]'.format(i_idx)
                i_idx += 1

                p['IDELAY_ONLY'] = random.randint(0, 1)
                p['DIFF_TERM'] = random.randint(0, 1)
                if not p['IDELAY_ONLY']:
                    p['owire'] = luts.get_next_input_net()
                else:
                    any_idelay = True
                    p['owire'] = 'idelay_{site}'.format(**p)

                p['DRIVE'] = None
                p['SLEW'] = None
                p['IBUF_LOW_PWR'] = random.randint(0, 1)

            elif p['type'] == 'OBUF':
                p['pad_wire'] = 'do[{}]'.format(o_idx)
                p['iwire'] = luts.get_next_output_net()
                if drives is not None:
                    p['DRIVE'] = random.choice(drives)
                else:
                    p['DRIVE'] = None
                p['SLEW'] = verilog.quote(random.choice(slews))

                o_idx += 1
            elif p['type'] == 'OBUFDS':
                p['pad_wire'] = 'do[{}]'.format(o_idx)
                o_idx += 1
                p['bpad_wire'] = 'do[{}]'.format(o_idx)
                o_idx += 1
                p['iwire'] = luts.get_next_output_net()
                if drives is not None:
                    p['DRIVE'] = random.choice(drives)
                else:
                    p['DRIVE'] = None
                p['SLEW'] = verilog.quote(random.choice(slews))
            elif p['type'] == 'OBUFTDS':
                p['pad_wire'] = 'do[{}]'.format(o_idx)
                o_idx += 1
                p['bpad_wire'] = 'do[{}]'.format(o_idx)
                o_idx += 1
                p['tristate_wire'] = random.choice(
                    ('0', luts.get_next_output_net()))
                p['iwire'] = luts.get_next_output_net()
                if drives is not None:
                    p['DRIVE'] = random.choice(drives)
                else:
                    p['DRIVE'] = None
                p['SLEW'] = verilog.quote(random.choice(slews))
            elif p['type'] == 'IOBUF_INTERMDISABLE':
                p['pad_wire'] = 'dio[{}]'.format(io_idx)
                p['iwire'] = luts.get_next_output_net()
                p['owire'] = luts.get_next_input_net()
                if drives is not None:
                    p['DRIVE'] = random.choice(drives)
                else:
                    p['DRIVE'] = None
                p['SLEW'] = verilog.quote(random.choice(slews))
                p['tristate_wire'] = random.choice(
                    ('0', luts.get_next_output_net()))
                p['ibufdisable_wire'] = random.choice(
                    ('0', luts.get_next_output_net()))
                p['intermdisable_wire'] = random.choice(
                    ('0', luts.get_next_output_net()))
                io_idx += 1

            if 'DRIVE' in p:
                if p['DRIVE'] is not None:
                    p['DRIVE_STR'] = '.DRIVE({}),'.format(p['DRIVE'])
                else:
                    p['DRIVE_STR'] = ''

            if 'SLEW' in p:
                p['SLEW_STR'] = ''
                if iostandard in only_diff_map:
                    p['SLEW'] = None
                elif p['DRIVE'] is not None:
                    p['SLEW_STR'] = '.SLEW({}),'.format(p['SLEW'])

            if p['type'] is not None:
                tile_params.append(
                    (
                        tile,
                        site,
                        p['pad_wire'],
                        iostandard_site,
                        p['DRIVE'],
                        verilog.unquote(p['SLEW']) if p['SLEW'] else None,
                        verilog.unquote(p['PULLTYPE']),
                        p['IN_TERM'] if 'IN_TERM' in p else None,
                    ))
            params['tiles'].append(p)

    write_params(tile_params)

    with open('iobank_vref.csv', 'w') as f:
        for iobank, vref in params['INTERNAL_VREF'].items():
            f.write('{},{}\n'.format(iobank, vref))

    print(
        '''
`define N_DI {n_di}
`define N_DO {n_do}
`define N_DIO {n_dio}

module top(input wire [`N_DI-1:0] di, output wire [`N_DO-1:0] do, inout wire [`N_DIO-1:0] dio);
    '''.format(n_di=i_idx, n_do=o_idx, n_dio=io_idx))

    if any_idelay:
        print('''
        (* KEEP, DONT_TOUCH *)
        IDELAYCTRL();''')

    # Always output a LUT6 to make placer happy.
    print('''
        (* KEEP, DONT_TOUCH *)
        LUT6 dummy_lut();''')

    for p in params['tiles']:
        if p['type'] is None:
            continue
        elif p['type'] == 'IBUF':
            print(
                '''
        wire idelay_{site};

        (* KEEP, DONT_TOUCH *)
        IBUF #(
            .IBUF_LOW_PWR({IBUF_LOW_PWR}),
            .IOSTANDARD({IOSTANDARD})
        ) ibuf_{site} (
            .I({pad_wire}),
            .O({owire})
            );'''.format(**p),
                file=connects)
            if p['IDELAY_ONLY']:
                print(
                    """
        (* KEEP, DONT_TOUCH *)
        IDELAYE2 idelay_site_{site} (
            .IDATAIN(idelay_{site})
            );""".format(**p),
                    file=connects)

        elif p['type'] == 'IBUFDS':
            print(
                '''
        wire idelay_{site};

        (* KEEP, DONT_TOUCH *)
        IBUFDS #(
            .IBUF_LOW_PWR({IBUF_LOW_PWR}),
            .DIFF_TERM({DIFF_TERM}),
            .IOSTANDARD({IOSTANDARD})
        ) ibuf_{site} (
            .I({pad_wire}),
            .IB({bpad_wire}),
            .O({owire})
            );'''.format(**p),
                file=connects)
            if p['IDELAY_ONLY']:
                print(
                    """
        (* KEEP, DONT_TOUCH *)
        IDELAYE2 idelay_site_{site} (
            .IDATAIN(idelay_{site})
            );""".format(**p),
                    file=connects)

        elif p['type'] == 'OBUF':
            print(
                '''
        (* KEEP, DONT_TOUCH *)
        OBUF #(
            {DRIVE_STR}
            {SLEW_STR}
            .IOSTANDARD({IOSTANDARD})
        ) obuf_{site} (
            .O({pad_wire}),
            .I({iwire})
            );'''.format(**p),
                file=connects)
        elif p['type'] == 'OBUFDS':
            print(
                '''
        (* KEEP, DONT_TOUCH *)
        OBUFDS #(
            {DRIVE_STR}
            {SLEW_STR}
            .IOSTANDARD({IOSTANDARD})
        ) obufds_{site} (
            .O({pad_wire}),
            .OB({bpad_wire}),
            .I({iwire})
            );'''.format(**p),
                file=connects)
        elif p['type'] == 'OBUFTDS':
            print(
                '''
        (* KEEP, DONT_TOUCH *)
        OBUFTDS #(
            {DRIVE_STR}
            {SLEW_STR}
            .IOSTANDARD({IOSTANDARD})
        ) obufds_{site} (
            .O({pad_wire}),
            .OB({bpad_wire}),
            .T({tristate_wire}),
            .I({iwire})
            );'''.format(**p),
                file=connects)
        elif p['type'] == 'IOBUF_INTERMDISABLE':
            print(
                '''
        (* KEEP, DONT_TOUCH *)
        IOBUF_INTERMDISABLE #(
            {DRIVE_STR}
            {SLEW_STR}
            .IOSTANDARD({IOSTANDARD})
        ) ibuf_{site} (
            .IO({pad_wire}),
            .I({iwire}),
            .O({owire}),
            .T({tristate_wire}),
            .IBUFDISABLE({ibufdisable_wire}),
            .INTERMDISABLE({intermdisable_wire})
            );'''.format(**p),
                file=connects)

    for l in luts.create_wires_and_luts():
        print(l)

    print(connects.getvalue())

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params, f, indent=2)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

ports = {
    "GTPE2_CHANNEL": [
        ("CFGRESET", 1),
        ("CLKRSVD0", 1),
        ("CLKRSVD1", 1),
        ("DMONFIFORESET", 1),
        ("DMONITORCLK", 1),
        ("DRPCLK", 1),
        ("DRPEN", 1),
        ("DRPWE", 1),
        ("EYESCANMODE", 1),
        ("EYESCANRESET", 1),
        ("EYESCANTRIGGER", 1),
        ("GTRESETSEL", 1),
        ("GTRXRESET", 1),
        ("GTTXRESET", 1),
        ("PMARSVDIN0", 1),
        ("PMARSVDIN1", 1),
        ("PMARSVDIN2", 1),
        ("PMARSVDIN3", 1),
        ("PMARSVDIN4", 1),
        ("RESETOVRD", 1),
        ("RX8B10BEN", 1),
        ("RXBUFRESET", 1),
        ("RXCDRFREQRESET", 1),
        ("RXCDRHOLD", 1),
        ("RXCDROVRDEN", 1),
        ("RXCDRRESET", 1),
        ("RXCDRRESETRSV", 1),
        ("RXCHBONDEN", 1),
        ("RXCHBONDMASTER", 1),
        ("RXCHBONDSLAVE", 1),
        ("RXCOMMADETEN", 1),
        ("RXDDIEN", 1),
        ("RXDFEXYDEN", 1),
        ("RXDLYBYPASS", 1),
        ("RXDLYEN", 1),
        ("RXDLYOVRDEN", 1),
        ("RXDLYSRESET", 1),
        ("RXGEARBOXSLIP", 1),
        ("RXLPMHFHOLD", 1),
        ("RXLPMHFOVRDEN", 1),
        ("RXLPMLFHOLD", 1),
        ("RXLPMLFOVRDEN", 1),
        ("RXLPMOSINTNTRLEN", 1),
        ("RXLPMRESET", 1),
        ("RXMCOMMAALIGNEN", 1),
        ("RXOOBRESET", 1),
        ("RXOSCALRESET", 1),
        ("RXOSHOLD", 1),
        ("RXOSINTEN", 1),
        ("RXOSINTHOLD", 1),
        ("RXOSINTNTRLEN", 1),
        ("RXOSINTOVRDEN", 1),
        ("RXOSINTPD", 1),
        ("RXOSINTSTROBE", 1),
        ("RXOSINTTESTOVRDEN", 1),
        ("RXOSOVRDEN", 1),
        ("RXPCOMMAALIGNEN", 1),
        ("RXPCSRESET", 1),
        ("RXPHALIGN", 1),
        ("RXPHALIGNEN", 1),
        ("RXPHDLYPD", 1),
        ("RXPHDLYRESET", 1),
        ("RXPHOVRDEN", 1),
        ("RXPMARESET", 1),
        ("RXPOLARITY", 1),
        ("RXPRBSCNTRESET", 1),
        ("RXRATEMODE", 1),
        ("RXSLIDE", 1),
        ("RXSYNCALLIN", 1),
        ("RXSYNCIN", 1),
        ("RXSYNCMODE", 1),
        ("RXUSERRDY", 1),
        ("RXUSRCLK2", 1),
        ("RXUSRCLK", 1),
        ("SETERRSTATUS", 1),
        ("SIGVALIDCLK", 1),
        ("TX8B10BEN", 1),
        ("TXCOMINIT", 1),
        ("TXCOMSAS", 1),
        ("TXCOMWAKE", 1),
        ("TXDEEMPH", 1),
        ("TXDETECTRX", 1),
        ("TXDIFFPD", 1),
        ("TXDLYBYPASS", 1),
        ("TXDLYEN", 1),
        ("TXDLYHOLD", 1),
        ("TXDLYOVRDEN", 1),
        ("TXDLYSRESET", 1),
        ("TXDLYUPDOWN", 1),
        ("TXELECIDLE", 1),
        ("TXINHIBIT", 1),
        ("TXPCSRESET", 1),
        ("TXPDELECIDLEMODE", 1),
        ("TXPHALIGN", 1),
        ("TXPHALIGNEN", 1),
        ("TXPHDLYPD", 1),
        ("TXPHDLYRESET", 1),
        ("TXPHDLYTSTCLK", 1),
        ("TXPHINIT", 1),
        ("TXPHOVRDEN", 1),
        ("TXPIPPMEN", 1),
        ("TXPIPPMOVRDEN", 1),
        ("TXPIPPMPD", 1),
        ("TXPIPPMSEL", 1),
        ("TXPISOPD", 1),
        ("TXPMARESET", 1),
        ("TXPOLARITY", 1),
        ("TXPOSTCURSORINV", 1),
        ("TXPRBSFORCEERR", 1),
        ("TXPRECURSORINV", 1),
        ("TXRATEMODE", 1),
        ("TXSTARTSEQ", 1),
        ("TXSWING", 1),
        ("TXSYNCALLIN", 1),
        ("TXSYNCIN", 1),
        ("TXSYNCMODE", 1),
        ("TXUSERRDY", 1),
        ("TXUSRCLK2", 1),
        ("TXUSRCLK", 1),
        ("RXADAPTSELTEST", 14),
        ("DRPDI", 16),
        ("GTRSVD", 16),
        ("PCSRSVDIN", 16),
        ("TSTIN", 20),
        ("RXELECIDLEMODE", 2),
        ("RXPD", 2),
        ("RXSYSCLKSEL", 2),
        ("TXPD", 2),
        ("TXSYSCLKSEL", 2),
        ("LOOPBACK", 3),
        ("RXCHBONDLEVEL", 3),
        ("RXOUTCLKSEL", 3),
        ("RXPRBSSEL", 3),
        ("RXRATE", 3),
        ("TXBUFDIFFCTRL", 3),
        ("TXHEADER", 3),
        ("TXMARGIN", 3),
        ("TXOUTCLKSEL", 3),
        ("TXPRBSSEL", 3),
        ("TXRATE", 3),
        ("TXDATA", 32),
        ("RXCHBONDI", 4),
        ("RXOSINTCFG", 4),
        ("RXOSINTID0", 4),
        ("TX8B10BBYPASS", 4),
        ("TXCHARDISPMODE", 4),
        ("TXCHARDISPVAL", 4),
        ("TXCHARISK", 4),
        ("TXDIFFCTRL", 4),
        ("TXPIPPMSTEPSIZE", 5),
        ("TXPOSTCURSOR", 5),
        ("TXPRECURSOR", 5),
        ("TXMAINCURSOR", 7),
        ("TXSEQUENCE", 7),
        ("DRPADDR", 9),
    ],
    "GTPE2_COMMON": [
        ("BGBYPASSB", 1),
        ("BGMONITORENB", 1),
        ("BGPDB", 1),
        ("BGRCALOVRDENB", 1),
        ("DRPCLK", 1),
        ("DRPEN", 1),
        ("DRPWE", 1),
        ("PLL0LOCKDETCLK", 1),
        ("PLL0LOCKEN", 1),
        ("PLL0PD", 1),
        ("PLL0RESET", 1),
        ("PLL1LOCKDETCLK", 1),
        ("PLL1LOCKEN", 1),
        ("PLL1PD", 1),
        ("PLL1RESET", 1),
        ("RCALENB", 1),
        ("DRPDI", 16),
        ("PLLRSVD1", 16),
        ("PLL0REFCLKSEL", 3),
        ("PLL1REFCLKSEL", 3),
        ("BGRCALOVRD", 5),
        ("PLLRSVD2", 5),
        ("DRPADDR", 8),
        ("PMARSVD", 8),
    ],
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
Canonicalize the Project X-Ray database files by sorting. The aim is to reduce
the diff output between runs to make it clearer what has changed.

DB Files
--------

DB files are sorted into "natural" ordering. This is generally the order that a
human would sort them in rather than how they sort as ASCII strings.

For example with tags, a sequence of ABC1 to ABC12 would have the ASCII sort
order of;

    ABC1
    ABC10
    ABC11
    ABC12
    ABC2
    ...
    ABC9

We instead sort them like the following;

    ABC1
    ABC2
    ...
    ABC9
    ABC10
    ABC11
    ABC12

For the segbit files, we sort the bit definitions ignoring any leading
exclamation mark. Doing this generally makes it much easier to see patterns in
the bit descriptions and you end up with output like the following for 1-hot
encoded choices,

    ABC.CHOICE1 22_15 !22_16 !22_17
    ABC.CHOICE2 !22_15 22_16 !22_17
    ABC.CHOICE3 !22_15 !22_16 22_17

JSON Files
----------

For the JSON files, we run them through Python's pretty printing module and
sort sets (lists where the order doesn't matter).

"""

import csv
import os
import random
import re
import sys

import json
import utils.xjson as xjson
import utils.cmp as cmp

from prjxray.util import OpenSafeFile

def split_all(s, chars):
    """Split on multiple character values.

    >>> split_all('a_b_c_d', '_. ')
    ['a', 'b', 'c', 'd']
    >>> split_all('a b c d', '_. ')
    ['a', 'b', 'c', 'd']
    >>> split_all('a.b.c.d', '_. ')
    ['a', 'b', 'c', 'd']
    >>> split_all('a_b.c d', '_. ')
    ['a', 'b', 'c', 'd']
    >>> split_all('a b_c.d', '_. ')
    ['a', 'b', 'c', 'd']
    """
    chars = list(chars)

    o = [s]
    while len(chars) > 0:
        c = chars.pop(0)

        n = []
        for i in o:
            n += i.split(c)

        o = n
    return o


NUM_REGEX = re.compile('^(.*?)([0-9]*)$')


def extract_num(i):
    """Extract number from a string to be sorted.

    >>> extract_num('BLAH123')
    ('BLAH', 123)
    >>> extract_num('123')
    123
    >>> extract_num('BLAH')
    'BLAH'
    >>> extract_num('')
    ''
    """
    g = NUM_REGEX.match(i).groups()
    if len(g[-1]) == 0:
        return i
    i = int(g[-1])
    if len(g[0]) == 0:
        return i
    else:
        return (g[0], i)


class bit(tuple):
    """Class representing a bit specifier.

    >>> a = bit.parse("02_12")
    >>> a
    bit(2, 12, True)
    >>> b = bit.parse("!02_03")
    >>> b
    bit(2, 3, False)
    >>> b == a
    False
    >>> b < a
    True
    >>> str(a)
    '02_12'
    >>> str(b)
    '!02_03'

    >>> bit.parseline("!30_04 !31_00 !31_01 31_02")
    [bit(30, 4, False), bit(31, 0, False), bit(31, 1, False), bit(31, 2, True)]

    >>> bit.parseline("31_02 !31_00 !31_01 !30_04")
    [bit(30, 4, False), bit(31, 0, False), bit(31, 1, False), bit(31, 2, True)]
    """

    @classmethod
    def parse(cls, s):
        mode = s[0] != '!'
        s = s.replace('!', '')
        assert '_' in s, s
        a, b = s.split('_', 1)
        assert '_' not in b, s
        return cls((extract_num(a), extract_num(b), mode))

    @classmethod
    def parseline(cls, s):
        bits = [cls.parse(b) for b in s.split(' ')]
        bits.sort()
        return bits

    def __repr__(self):
        return "bit" + tuple.__repr__(self)

    def __str__(self):
        s = self
        return "{}{:02d}_{:02d}".format(['!', ''][s[2]], s[0], s[1])


def convert_bit(i):
    """Convert a bit pattern into sortable form.

    >>> convert_bit("02_12")
    bit(2, 12, True)
    >>> convert_bit("!02_12")
    bit(2, 12, False)
    >>> convert_bit("!02_02")
    bit(2, 2, False)
    >>> convert_bit("always")
    'always'
    """
    if '_' not in i:
        return i
    return bit.parse(i)


def segbit_line_sort_bits(l):
    """Sort the bit section of a segbit line.

    >>> segbit_line_sort_bits("A !28_35 !27_39 27_37")
    'A 27_37 !27_39 !28_35'

    >>> segbit_line_sort_bits("B !28_35 !27_39 !27_37")
    'B !27_37 !27_39 !28_35'

    >>> segbit_line_sort_bits("C 28_35 00_00 !27_37")
    'C 00_00 !27_37 28_35'

    """
    tag, *segbits = l.split()

    segbits = [bit.parse(b) for b in segbits]
    segbits.sort()

    return "{} {}".format(tag, " ".join(str(s) for s in segbits))


def sortable_tag(t):
    """
    >>> sortable_tag("CLBLL_L.CLBLL_L_A.CLBLL_L_A1")
    ('CLBLL', 'L', 'CLBLL', 'L', 'A', 'CLBLL', 'L', ('A', 1))

    >>> sortable_tag("CLBLL_L.CLBLL_LOGIC_OUTS23.CLBLL_LL_DMUX")
    ('CLBLL', 'L', 'CLBLL', 'LOGIC', ('OUTS', 23), 'CLBLL', 'LL', 'DMUX')

    >>> sortable_tag("BRAM_L.RAMB18_Y0.INIT_B[9]")
    ('BRAM', 'L', ('RAMB', 18), ('Y', 0), 'INIT', 'B', 9)

    >>> sortable_tag("BRAM_L.RAMB18_Y0.READ_WIDTH_A_18")
    ('BRAM', 'L', ('RAMB', 18), ('Y', 0), 'READ', 'WIDTH', 'A', 18)
    """
    return tuple(extract_num(i) for i in split_all(t, '_.[]') if i != '')


def sortable_line_from_mask(l):
    """Convert a line in a mask_XXXX.db file to something sortable.

    Example lines from mask_XXX.db file
    >>> a, b = sortable_line_from_mask("bit 00_00")
    >>> a
    bit(0, 0, True)
    >>> b
    'bit 00_00'

    >>> a, b = sortable_line_from_mask("bit 09_39")
    >>> a
    bit(9, 39, True)
    >>> b
    'bit 09_39'
    """
    tag, b = l.split(' ', 1)
    assert tag == 'bit', tag
    return bit.parse(b), l


def sortable_line_from_ppips(l):
    """Convert a line in a ppips_XXX.db file to something sortable.

    Example lines from ppips_XXX.db file
    >>> a, b = sortable_line_from_ppips("CLBLL_L.CLBLL_L_A.CLBLL_L_A1 hint")
    >>> a
    (('CLBLL', 'L', 'CLBLL', 'L', 'A', 'CLBLL', 'L', ('A', 1)), 'hint')
    >>> b
    'CLBLL_L.CLBLL_L_A.CLBLL_L_A1 hint'

    >>> a, b = sortable_line_from_ppips("CLBLL_L.CLBLL_LOGIC_OUTS23.CLBLL_LL_DMUX always")
    >>> a
    (('CLBLL', 'L', 'CLBLL', 'LOGIC', ('OUTS', 23), 'CLBLL', 'LL', 'DMUX'), 'always')
    >>> b
    'CLBLL_L.CLBLL_LOGIC_OUTS23.CLBLL_LL_DMUX always'
    """
    assert ' ' in l, repr(l)
    tag, ptype = l.split(' ', 1)
    tag = sortable_tag(tag)
    return (tag, ptype), l


def sortable_line_from_segbits(l):
    """Convert a line in segbits_XXX.db file to something sortable.

    >>> (tag, bits), b = sortable_line_from_segbits("BRAM_L.RAMB18_Y0.INIT_B[9] 27_15")
    >>> tag
    ('BRAM', 'L', ('RAMB', 18), ('Y', 0), 'INIT', 'B', 9)
    >>> bits
    (bit(27, 15, True),)
    >>> b
    'BRAM_L.RAMB18_Y0.INIT_B[9] 27_15'

    >>> (tag, bits), b = sortable_line_from_segbits("BRAM_L.RAMB18_Y0.READ_WIDTH_A_18 !28_35 !27_39 27_37")
    >>> tag
    ('BRAM', 'L', ('RAMB', 18), ('Y', 0), 'READ', 'WIDTH', 'A', 18)
    >>> bits
    (bit(27, 37, True), bit(27, 39, False), bit(28, 35, False))
    >>> b
    'BRAM_L.RAMB18_Y0.READ_WIDTH_A_18 27_37 !27_39 !28_35'
    """
    tag, sbit = l.split(' ', 1)
    tag = sortable_tag(tag)

    bits = bit.parseline(sbit)

    l = segbit_line_sort_bits(l)
    return (tag, tuple(bits)), l


def sortable_line_from_origin_segbits(l):
    tag, origin, sbit = l.split(' ', 2)
    tag = sortable_tag(tag)

    bits = bit.parseline(sbit)

    return (tag, tuple(bits)), l


def sort_db(pathname):
    """Sort a XXX.db file."""
    filename = os.path.split(pathname)[-1]
    if filename.startswith('segbits_'):
        if 'origin_info' in filename:
            sortable_line_from_dbfile = sortable_line_from_origin_segbits
        else:
            sortable_line_from_dbfile = sortable_line_from_segbits
    elif 'origin_info' in filename:
        return False
    elif filename.startswith('ppips_'):
        sortable_line_from_dbfile = sortable_line_from_ppips
    elif filename.startswith('grid-'):
        sortable_line_from_dbfile = sortable_line_from_ppips
    elif filename.startswith('mask_'):
        sortable_line_from_dbfile = sortable_line_from_mask
    else:
        return False

    with OpenSafeFile(pathname) as f:
        lines = f.readlines()

    tosort = []
    for l in lines:
        l = l.strip()
        if not l:
            continue
        tosort.append(sortable_line_from_dbfile(l))

    tosort.sort(key=cmp.cmp_key)

    with open(pathname, 'w') as f:
        for _, l in tosort:
            f.write(l)
            f.write('\n')

    return True


def sort_csv(pathname):
    rows = []
    fields = []
    delimiter = None
    with open(pathname, newline='') as f:
        if pathname.endswith('.csv'):
            delimiter = ','
        elif pathname.endswith('.txt'):
            delimiter = ' '
        reader = csv.DictReader(f, delimiter=delimiter)
        fields.extend(reader.fieldnames)
        rows.extend(reader)
        del reader

    def sort_key(r):
        v = []
        for field in fields:
            v.append(sortable_tag(r[field]))
        return tuple(v)

    rows.sort(key=sort_key)

    with open(pathname, 'w', newline='') as f:
        writer = csv.DictWriter(
            f, fields, delimiter=delimiter, lineterminator='\n')
        writer.writeheader()
        writer.writerows(rows)

    return True


def sort_json(filename):
    """Sort a XXX.json file."""

    try:
        with OpenSafeFile(filename) as f:
            d = json.load(f)
    except json.JSONDecodeError as e:
        print(e)
        return False

    with OpenSafeFile(filename, 'w') as f:
        xjson.pprint(f, d)

    return True


def sort_db_text(n):
    rows = []
    with OpenSafeFile(n) as f:
        for l in f:
            rows.append(([extract_num(s) for s in l.split()], l))

    rows.sort(key=lambda i: i[0])

    with OpenSafeFile(n, 'w') as f:
        for l in rows:
            f.write(l[-1])

    return True


def sort_file(n):

    assert os.path.exists(n)

    base, ext = os.path.splitext(n)
    dirname, base = os.path.split(base)

    # Leave db files with fuzzer of origin untouched
    if "origin_info" in n and not base.startswith('segbits'):
        print("Ignoring     file {:45s}".format(n), flush=True)
        return

    if ext == '.db':
        print("Sorting DB   file {:45s}".format(n), end=" ", flush=True)
        x = sort_db(n)
    elif ext == '.json':
        print("Sorting JSON file {:45s}".format(n), end=" ", flush=True)
        x = sort_json(n)
    elif ext in ('.csv', '.txt'):
        if n.endswith('-db.txt'):
            print("Sorting txt  file {:45s}".format(n), end=" ", flush=True)
            x = sort_db_text(n)
        else:
            print("Sorting CSV  file {:45s}".format(n), end=" ", flush=True)
            x = sort_csv(n)
    else:
        print("Ignoring     file {:45s}".format(n), end=" ", flush=True)
        x = True
    if x:
        print(".. success.")
    else:
        print(".. failed.")


def sort_dir(dirname):
    for n in sorted(os.listdir(dirname)):
        n = os.path.join(dirname, n)
        if os.path.isdir(n):
            print("Entering     dir  {:45s}".format(n), flush=True)
            sort_dir(n)
            continue
        elif not os.path.isfile(n):
            print("Ignoring non-file {:45s}".format(n), flush=True)
            continue

        sort_file(n)


def main(argv):
    if argv[1:]:
        for n in argv[1:]:
            sort_file(n)
    else:
        sort_dir('.')
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray import util


class StateGen(object):
    """ Manages fuzzer state generation across multiple sites.

    sites - List of sites.
    states_per_site - Maximum number of states used per site.

    If states_per_site is too small, next_state may throw AssertionError.

    StateGen should be used as a iterator for the sites given.  Call next_state
    within each site output loop.  Once 'next' is called on StateGen, StateGen
    will advance state output to the correct position, even if next_state was
    called less than states_per_site.

    """

    def __init__(self, sites, states_per_site):
        self.sites = sites
        self.states_per_site = states_per_site
        self.curr_site_idx = 0
        self.curr_state = None
        self.states = None
        self.curr_site = None

    def __iter__(self):
        assert self.curr_state is None
        assert self.states is None
        assert self.curr_state is None

        self.curr_site_idx = 0
        self.curr_state = None
        self.states = util.gen_fuzz_states(
            len(self.sites) * self.states_per_site)
        self.curr_site = iter(self.sites)
        return self

    def __next__(self):
        next_site = next(self.curr_site)
        self.curr_site_idx += 1

        if self.curr_state is not None:
            while self.curr_state < self.states_per_site:
                self.next_state()

            assert self.curr_state == self.states_per_site, self.curr_state

        self.curr_state = 0

        return next_site

    def next_state(self):
        """ Returns next state within site.

        Should only be called states_per_site for each site.
        """
        self.curr_state += 1

        try:
            state = next(self.states)
        except StopIteration:
            assert False, "Insufficent states, at state {} for site {}".format(
                self.curr_state, self.curr_site_idx)

        return state



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray.util import OpenSafeFile

class Bitfilter(object):
    def __init__(
            self, frames_to_include=None, frames_to_exclude=[],
            bits_to_exclude=[]):
        self.frames_to_include = frames_to_include
        self.frames_to_exclude = frames_to_exclude
        self.bits_to_exclude = bits_to_exclude

    def filter(self, frame, bit):
        if self.frames_to_include is not None:
            if frame in self.frames_to_include:
                return True

        if frame in self.frames_to_exclude:
            return False

        if (frame, bit) in self.bits_to_exclude:
            return False

        return True


BITFILTERS = {
    ('artix7', 'INT'):
    Bitfilter(
        frames_to_exclude=[
            30,
            31,
        ],
        bits_to_exclude=[
            #
            (0, 36)
        ]),
}


def get_bitfilter(part, tile):
    """ Returns bitfilter for specified part and tile.

    Either returns bitfilter to specified part and tile type, or the default
    bitfilter, which includes all bits.
    """
    key = (part, tile)
    if key in BITFILTERS:
        return BITFILTERS[key].filter
    else:
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

# FIXME: getting two bits
# 00_40 31_46
# Can we find instance where they are not aliased?
WA7USED = 0

from prjxray.segmaker import Segmaker

segmk = Segmaker("design.bits")

print("Loading tags")
'''
module,loc,c31,b31,a31
my_NDI1MUX_NI_NMC31,SLICE_X12Y100,1,1,0
my_NDI1MUX_NI_NMC31,SLICE_X12Y101,1,1,1
my_NDI1MUX_NI_NMC31,SLICE_X12Y102,1,1,1
'''
f = open('params.csv', 'r')
f.readline()
for l in f:
    l = l.strip()
    module, loc, c31, b31, a31 = l.split(',')
    c31 = int(c31)
    b31 = int(b31)
    a31 = int(a31)
    segmk.add_site_tag(loc, "ALUT.DI1MUX.AI", 1 ^ a31)
    segmk.add_site_tag(loc, "ALUT.DI1MUX.BDI1_BMC31", a31)
    segmk.add_site_tag(loc, "BLUT.DI1MUX.BI", 1 ^ b31)
    segmk.add_site_tag(loc, "BLUT.DI1MUX.DI_CMC31", b31)
    segmk.add_site_tag(loc, "CLUT.DI1MUX.CI", 1 ^ c31)
    segmk.add_site_tag(loc, "CLUT.DI1MUX.DI_DMC31", c31)

segmk.compile()
segmk.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray import util
import os
import json
'''
Local utils script to hold shared code of the 005-tilegrid fuzzer scripts
'''


class TileFrames:
    """
    Class for getting the number of frames used for configuring a tile
    with the specified baseaddress using the information from the part's json file
    """

    def __init__(self):
        self.tile_address_to_frames = dict()

    def get_baseaddress(self, region, bus, row, column):
        assert bus == 'BLOCK_RAM' or bus == 'CLB_IO_CLK', 'Incorrect block type'
        address = (row << 17) + (column << 7) + (
            (1 << 22) if region == 'bottom' else 0) + (
                (1 << 23) if bus == 'BLOCK_RAM' else 0)
        return address

    def initialize_address_to_frames(self):
        with open(os.path.join(os.getenv('XRAY_FAMILY_DIR'),
                               os.getenv('XRAY_PART'), 'part.json')) as pf:
            part_json = json.load(pf)
        for clock_region, rows in part_json['global_clock_regions'].items():
            for row, buses in rows['rows'].items():
                for bus, columns in buses['configuration_buses'].items():
                    for column, frames in columns[
                            'configuration_columns'].items():
                        address = self.get_baseaddress(
                            clock_region, bus, int(row), int(column))
                        assert address not in self.tile_address_to_frames
                        self.tile_address_to_frames[address] = frames[
                            'frame_count']

    def get_tile_frames(self, baseaddress):
        if len(self.tile_address_to_frames) == 0:
            self.initialize_address_to_frames()
        assert baseaddress in self.tile_address_to_frames, "Base address not found in the part's json file"
        return self.tile_address_to_frames[baseaddress]


def get_entry(tile_type, block_type):
    """ Get frames and words for a given tile_type (e.g. CLBLL) and block_type (CLB_IO_CLK, BLOCK_RAM, etc). """
    return {
        # (tile_type, block_type): (frames, words, height)
        ("CLBLL", "CLB_IO_CLK"): (36, 2, None),
        ("CLBLM", "CLB_IO_CLK"): (36, 2, None),
        ("HCLK", "CLB_IO_CLK"): (26, 1, None),
        ("INT", "CLB_IO_CLK"): (28, 2, None),
        ("BRAM", "CLB_IO_CLK"): (28, 10, None),
        ("BRAM", "BLOCK_RAM"): (128, 10, None),
        ("DSP", "CLB_IO_CLK"): (28, 2, None),
    }.get((tile_type, block_type), None)


def get_int_params():
    int_frames, int_words, _ = get_entry('INT', 'CLB_IO_CLK')
    return int_frames, int_words


def add_tile_bits(
        tile_name,
        tile_db,
        baseaddr,
        offset,
        frames,
        words,
        tile_frames,
        verbose=False):
    '''
    Record data structure geometry for the given tile baseaddr
    For most tiles there is only one baseaddr, but some like BRAM have multiple
    Notes on multiple block types:
    https://github.com/SymbiFlow/prjxray/issues/145
    '''
    bits = tile_db['bits']
    block_type = util.addr2btype(baseaddr)

    # Extract the information about the maximal number of frames from the part's json
    max_frames = tile_frames.get_tile_frames(baseaddr)
    if frames > max_frames:
        print(
            "Warning: The number of frames for base address {} specified for the tile {} ({}) exceeds the maximum allowed value ({}). Falling back to the maximum value."
            .format(hex(baseaddr), tile_name, frames, max_frames))
        frames = max_frames
    # If frames count is None then use the maximum
    if frames is None:
        frames = max_frames

    assert offset <= 100, (tile_name, offset)
    # Few rare cases at X=0 for double width tiles split in half => small negative offset
    assert offset >= 0 or "IOB" in tile_name or "GTX_INT_INTERFACE" in tile_name, (
        tile_name, hex(baseaddr), offset)
    assert 1 <= words <= 101, words
    assert offset + words <= 101, (
        tile_name, offset + words, offset, words, block_type)

    baseaddr_str = '0x%08X' % baseaddr
    block = bits.get(block_type, None)
    if block is not None:
        verbose and print(
            "%s: existing defintion for %s" % (tile_name, block_type))
        assert block["baseaddr"] == baseaddr_str
        assert block["frames"] == frames, (block, frames)
        assert block["offset"] == offset, "%s; orig offset %s, new %s" % (
            tile_name, block["offset"], offset)
        assert block["words"] == words
        return
    block = bits.setdefault(block_type, {})

    # FDRI address
    block["baseaddr"] = baseaddr_str
    # Number of frames this entry is sretched across
    # that is the following FDRI addresses are used: range(baseaddr, baseaddr + frames)
    block["frames"] = frames

    # Index of first word used within each frame
    block["offset"] = offset

    # Number of words used by tile.
    block["words"] = words



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import itertools
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_fifos():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['IN_FIFO']:
                if gridinfo.tile_type[-1] == 'L':
                    int_grid_x = loc.grid_x + 2
                    int_tile_type = 'INT_L'
                else:
                    int_grid_x = loc.grid_x - 2
                    int_tile_type = 'INT_R'

                int_tile_locs = [
                    (int_grid_x, loc.grid_y + idx - 5) for idx in range(12)
                ]

                int_tiles = []
                for int_tile_loc in int_tile_locs:
                    int_gridinfo = grid.gridinfo_at_loc(int_tile_loc)
                    assert int_gridinfo.tile_type == int_tile_type, (
                        int_tile_loc, int_gridinfo.tile_type, int_tile_type)

                    int_tiles.append(grid.tilename_at_loc(int_tile_loc))

                yield site_name, int_tiles


def write_params(params):
    pinstr = 'tile,val\n'
    for tile, (val, ) in sorted(params.items()):
        pinstr += '%s,%s\n' % (tile, val)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')

    params = {}

    sites = list(gen_fifos())
    N_INT = 12

    fuzz_iter = iter(util.gen_fuzz_states(len(sites) * N_INT))
    for site, int_tiles in sites:
        assert len(int_tiles) == N_INT
        int_tiles.reverse()

        bits = itertools.islice(fuzz_iter, N_INT)

        assigns = []

        # CMT_FIFO mux usage is regular with IMUX_L40 and IMUX_L42.
        #
        # INT[idx].IMUX_L40 = IN.D{idx}[1]
        # INT[idx].IMUX_L42 = IN.D{idx}[3]
        CHANNEL = [0, 1, 2, 3, 4, 5, 5, 6, 6, 7, 8, 9]
        HOLD_BIT_0 = [1, 1, 1, 1, 1, 1, 5, 5, 1, 1, 1, 1]
        TOGGLE_BIT = [3, 3, 3, 3, 3, 3, 7, 7, 3, 3, 3, 3]
        #             0  1  2  3  4  5  6  7  8  9
        WIDTH = [4, 4, 4, 4, 4, 8, 8, 4, 4, 4]

        bits_set = set()

        for idx, (int_tile, bit) in enumerate(zip(int_tiles, bits)):
            bits_set.add((CHANNEL[idx], HOLD_BIT_0[idx]))
            bits_set.add((CHANNEL[idx], TOGGLE_BIT[idx]))

            assigns.append('            // {}'.format(int_tile))
            assigns.append(
                '            assign {site}_in_d{channel}[{bit}] = 0;'.format(
                    site=site,
                    channel=CHANNEL[idx],
                    bit=HOLD_BIT_0[idx],
                ))
            assigns.append(
                '            assign {site}_in_d{channel}[{bit}] = {toggle_bit};'
                .format(
                    site=site,
                    channel=CHANNEL[idx],
                    bit=TOGGLE_BIT[idx],
                    toggle_bit=bit,
                ))
            params[int_tile] = (bit, )

            assigns.append('')

        for channel, width in enumerate(WIDTH):
            for bit in range(width):
                if (channel, bit) not in bits_set:
                    assigns.append(
                        '            assign {site}_in_d{channel}[{bit}] = 1;'.
                        format(
                            site=site,
                            channel=channel,
                            bit=bit,
                        ))

        print(
            '''
            wire [3:0] {site}_in_d0;
            wire [3:0] {site}_in_d1;
            wire [3:0] {site}_in_d2;
            wire [3:0] {site}_in_d3;
            wire [3:0] {site}_in_d4;
            wire [7:0] {site}_in_d5;
            wire [7:0] {site}_in_d6;
            wire [3:0] {site}_in_d7;
            wire [3:0] {site}_in_d8;
            wire [3:0] {site}_in_d9;

{assign_statements}

            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            IN_FIFO fifo_{site} (
                .D0({site}_in_d0),
                .D1({site}_in_d1),
                .D2({site}_in_d2),
                .D3({site}_in_d3),
                .D4({site}_in_d4),
                .D5({site}_in_d5),
                .D6({site}_in_d6),
                .D7({site}_in_d7),
                .D8({site}_in_d8),
                .D9({site}_in_d9)
                );
'''.format(
                site=site,
                assign_statements='\n'.join(assigns),
            ))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
import json
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile)
        gridinfo = grid.gridinfo_at_loc(loc)
        if gridinfo.tile_type in ['DSP_L', 'DSP_R']:
            for site in sorted(gridinfo.sites.keys()):
                if gridinfo.sites[site] == 'DSP48E1':
                    yield tile, site


def fuzz(*args):
    if len(args) == 1 and isinstance(args[0], int):
        # Argument indicates that we should generate a random integer with
        # args[0] number of bits.
        return random.getrandbits(args[0])
    else:
        # Otherwise make a random choice
        return random.choice(*args)


def run():
    verilog.top_harness(48, 48)

    print('module roi(input clk, input [47:0] din, output [47:0] dout);')

    data = {}
    data['instances'] = []

    sites = list(gen_sites())

    for i, (tile, site) in enumerate(sites):
        synthesis = '(* KEEP, DONT_TOUCH, LOC = "%s" *)' % (site)
        module = 'DSP48E1'
        instance = 'INST_%s' % (site)
        ports = {}
        params = {}

        ports['A'] = '{30{1\'b1}}'
        ports['ACIN'] = '{30{1\'b1}}'
        ports['ACOUT'] = '30\'b0'
        ports['ALUMODE'] = 'din[3:0]'
        ports['B'] = '{18{1\'b1}}'
        ports['BCIN'] = '{18{1\'b1}}'
        ports['BCOUT'] = '18\'b0'
        ports['C'] = '{48{1\'b1}}'
        ports['CARRYCASCIN'] = '1\'b1'
        ports['CARRYCASCOUT'] = '1\'b0'
        ports['CARRYIN'] = 'din[4]'
        ports['CARRYINSEL'] = '3\'b000'
        ports['CARRYOUT'] = '4\'b0'
        ports['CEA1'] = '1\'b1'
        ports['CEA2'] = '1\'b1'
        ports['CEAD'] = '1\'b1'
        ports['CEALUMODE'] = '1\'b1'
        ports['CEB1'] = '1\'b1'
        ports['CEB2'] = '1\'b1'
        ports['CEC'] = '1\'b1'
        ports['CECARRYIN'] = '1\'b1'
        ports['CECTRL'] = '1\'b1'
        ports['CED'] = '1\'b1'
        ports['CEINMODE'] = '1\'b1'
        ports['CEM'] = '1\'b1'
        ports['CEP'] = '1\'b1'
        ports['CLK'] = 'clk'
        ports['D'] = '{25{1\'b1}}'
        ports['INMODE'] = 'din[9:5]'
        #ports['MULTISIGNIN'] = '1\'b1'
        #ports['MULTISIGNOUT'] = '1\'b0'
        ports['OPMODE'] = 'din[16:10]'
        ports['OVERFLOW'] = '1\'b0'
        ports['P'] = '48\'b0'
        ports['PATTERNBDETECT'] = '1\'b0'
        ports['PATTERNDETECT'] = '1\'b0'
        ports['PCIN'] = '{48{1\'b1}}'
        ports['PCOUT'] = '48\'b0'
        ports['RSTA'] = '1\'b1'
        ports['RSTALLCARRYIN'] = '1\'b1'
        ports['RSTALUMODE'] = '1\'b1'
        ports['RSTB'] = '1\'b1'
        ports['RSTC'] = '1\'b1'
        ports['RSTCTRL'] = '1\'b1'
        ports['RSTD'] = '1\'b1'
        ports['RSTINMODE'] = '1\'b1'
        ports['RSTM'] = '1\'b1'
        ports['RSTP'] = '1\'b1'
        ports['UNDERFLOW'] = '1\'b0'

        params['ADREG'] = fuzz((0, 1))
        params['ALUMODEREG'] = fuzz((0, 1))
        params['AREG'] = fuzz((0, 1, 2))
        if params['AREG'] == 0 or params['AREG'] == 1:
            params['ACASCREG'] = params['AREG']
        else:
            params['ACASCREG'] = fuzz((1, 2))
        params['BREG'] = fuzz((0, 1, 2))
        if params['BREG'] == 0 or params['BREG'] == 1:
            params['BCASCREG'] = params['BREG']
        else:
            params['BCASCREG'] = fuzz((1, 2))
        params['CARRYINREG'] = fuzz((0, 1))
        params['CARRYINSELREG'] = fuzz((0, 1))
        params['CREG'] = fuzz((0, 1))
        params['DREG'] = fuzz((0, 1))
        params['INMODEREG'] = fuzz((0, 1))
        params['OPMODEREG'] = fuzz((0, 1))
        params['PREG'] = fuzz((0, 1))
        params['A_INPUT'] = verilog.quote(fuzz(('DIRECT', 'CASCADE')))
        params['B_INPUT'] = verilog.quote(fuzz(('DIRECT', 'CASCADE')))
        params['USE_DPORT'] = verilog.quote(fuzz(('TRUE', 'FALSE')))
        params['USE_SIMD'] = verilog.quote(fuzz(('ONE48', 'TWO24', 'FOUR12')))
        params['USE_MULT'] = verilog.quote(
            'NONE' if params['USE_SIMD'] != verilog.quote('ONE48') else fuzz(
                ('NONE', 'MULTIPLY', 'DYNAMIC')))
        params['MREG'] = 0 if params['USE_MULT'] == verilog.quote(
            'NONE') else fuzz((0, 1))
        params['AUTORESET_PATDET'] = verilog.quote(
            fuzz(('NO_RESET', 'RESET_MATCH', 'RESET_NOT_MATCH')))
        params['MASK'] = '48\'d%s' % fuzz(48)
        params['PATTERN'] = '48\'d%s' % fuzz(48)
        params['SEL_MASK'] = verilog.quote(
            fuzz(('MASK', 'C', 'ROUNDING_MODE1', 'ROUNDING_MODE2')))
        params['USE_PATTERN_DETECT'] = verilog.quote(
            fuzz(('NO_PATDET', 'PATDET')))

        params['IS_ALUMODE_INVERTED'] = fuzz(4)
        params['IS_CARRYIN_INVERTED'] = fuzz((0, 1))
        params['IS_CLK_INVERTED'] = fuzz((0, 1))
        params['IS_INMODE_INVERTED'] = fuzz(5)
        params['IS_OPMODE_INVERTED'] = fuzz(7)

        verilog.instance(synthesis + ' ' + module, instance, ports, params)

        params['TILE'] = tile
        params['SITE'] = site

        data['instances'].append(params)

    with open('params.json', 'w') as fp:
        json.dump(data, fp)

    print("endmodule")


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from __future__ import print_function

import fasm
import argparse
import json
import os
import os.path
import csv

from collections import defaultdict

from prjxray import fasm_assembler, util
from prjxray.db import Database
from prjxray.roi import Roi
from prjxray.util import OpenSafeFile

import sys


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


class FASMSyntaxError(SyntaxError):
    pass


def dump_frames_verbose(frames):
    print()
    print("Frames: %d" % len(frames))
    for addr in sorted(frames.keys()):
        words = frames[addr]
        print(
            '0x%08X ' % addr + ', '.join(['0x%08X' % w for w in words]) +
            '...')


def dump_frames_sparse(frames):
    print()
    print("Frames: %d" % len(frames))
    for addr in sorted(frames.keys()):
        words = frames[addr]

        # Skip frames without filled words
        for w in words:
            if w:
                break
        else:
            continue

        print('Frame @ 0x%08X' % addr)
        for i, w in enumerate(words):
            if w:
                print('  % 3d: 0x%08X' % (i, w))


def dump_frm(f, frames):
    '''Write a .frm file given a list of frames, each containing a list of 101 32 bit words'''
    for addr in sorted(frames.keys()):
        words = frames[addr]
        f.write(
            '0x%08X ' % addr + ','.join(['0x%08X' % w for w in words]) + '\n')


def find_pudc_b(db):
    """ Find PUDC_B pin func in grid, and return the tile and site prefix.

    The PUDC_B pin is a special 7-series pin that controls unused pin pullup.

    If the PUDC_B is unused, it is configured as an input with a PULLUP.

    """
    grid = db.grid()

    pudc_b_tile_site = None
    for tile in grid.tiles():
        gridinfo = grid.gridinfo_at_tilename(tile)

        for site, pin_function in gridinfo.pin_functions.items():
            if 'PUDC_B' in pin_function:
                assert pudc_b_tile_site == None, (
                    pudc_b_tile_site, (tile, site))
                iob_y = int(site[-1]) % 2

                pudc_b_tile_site = (tile, 'IOB_Y{}'.format(iob_y))

    return pudc_b_tile_site


def get_iob_sites(db, tile_name):
    """
    Yields prjxray site names for given IOB tile name
    """
    grid = db.grid()
    gridinfo = grid.gridinfo_at_tilename(tile_name)

    for site in gridinfo.sites:
        site_y = int(site[-1]) % 2
        yield "IOB_Y{}".format(site_y)


def run(
        db_root,
        part,
        filename_in,
        f_out,
        sparse=False,
        roi=None,
        debug=False,
        emit_pudc_b_pullup=False):
    db = Database(db_root, part)
    assembler = fasm_assembler.FasmAssembler(db)

    set_features = set()

    def feature_callback(feature):
        set_features.add(feature)

    assembler.set_feature_callback(feature_callback)

    # Build mapping of tile to IO bank
    tile_to_bank = {}
    bank_to_tile = defaultdict(lambda: set())

    if part is not None:
        with OpenSafeFile(os.path.join(db_root, part, "package_pins.csv"), "r") as fp:
            reader = csv.DictReader(fp)
            package_pins = [l for l in reader]

        with OpenSafeFile(os.path.join(db_root, part, "part.json"), "r") as fp:
            part_data = json.load(fp)

        for bank, loc in part_data["iobanks"].items():
            tile = "HCLK_IOI3_" + loc
            bank_to_tile[bank].add(tile)
            tile_to_bank[tile] = bank

        for pin in package_pins:
            bank_to_tile[pin["bank"]].add(pin["tile"])
            tile_to_bank[pin["tile"]] = pin["bank"]

    if emit_pudc_b_pullup:
        pudc_b_in_use = False
        pudc_b_tile_site = find_pudc_b(db)

        def check_for_pudc_b(set_feature):
            feature_callback(set_feature)
            parts = set_feature.feature.split('.')

            if parts[0] == pudc_b_tile_site[0] and parts[
                    1] == pudc_b_tile_site[1]:
                nonlocal pudc_b_in_use
                pudc_b_in_use = True

        if pudc_b_tile_site is not None:
            assembler.set_feature_callback(check_for_pudc_b)

    extra_features = []
    if roi:
        with OpenSafeFile(roi) as f:
            roi_j = json.load(f)
        x1 = roi_j['info']['GRID_X_MIN']
        x2 = roi_j['info']['GRID_X_MAX']
        y1 = roi_j['info']['GRID_Y_MIN']
        y2 = roi_j['info']['GRID_Y_MAX']

        assembler.mark_roi_frames(Roi(db=db, x1=x1, x2=x2, y1=y1, y2=y2))

        if 'required_features' in roi_j:
            extra_features = list(
                fasm.parse_fasm_string('\n'.join(roi_j['required_features'])))

    # Get required extra features for the part
    required_features = db.get_required_fasm_features(part)
    extra_features += list(
        fasm.parse_fasm_string('\n'.join(required_features)))

    assembler.parse_fasm_filename(filename_in, extra_features=extra_features)

    if emit_pudc_b_pullup and not pudc_b_in_use and pudc_b_tile_site is not None:
        # Enable IN-only and PULLUP on PUDC_B IOB.
        #
        # TODO: The following FASM string only works on Artix 50T and Zynq 10
        # fabrics.  It is known to be wrong for the K70T fabric, but it is
        # unclear how to know which IOSTANDARD to use.
        missing_features = []
        for line in fasm.parse_fasm_string("""
{tile}.{site}.LVCMOS12_LVCMOS15_LVCMOS18_LVCMOS25_LVCMOS33_LVTTL_SSTL135_SSTL15.IN_ONLY
{tile}.{site}.LVCMOS25_LVCMOS33_LVTTL.IN
{tile}.{site}.PULLTYPE.PULLUP
""".format(
                tile=pudc_b_tile_site[0],
                site=pudc_b_tile_site[1],
        )):
            assembler.add_fasm_line(line, missing_features)

        if missing_features:
            raise fasm_assembler.FasmLookupError('\n'.join(missing_features))

    if part is not None:
        # Make a set of all used IOB tiles and sites. Look for the "STEPDOWN"
        # feature. If one is set for an IOB then set it for all other IOBs of
        # the same bank.
        stepdown_tags = defaultdict(lambda: set())
        stepdown_banks = set()
        used_iob_sites = set()

        for set_feature in set_features:
            if set_feature.value == 0:
                continue

            feature = set_feature.feature
            parts = feature.split(".")
            if len(parts) >= 3:
                tile, site, tag = feature.split(".", maxsplit=2)
                if "IOB33" in tile:
                    used_iob_sites.add((
                        tile,
                        site,
                    ))

                # Store STEPDOWN related tags.
                if "STEPDOWN" in tag:
                    bank = tile_to_bank[tile]
                    stepdown_banks.add(bank)
                    stepdown_tags[bank].add(tag)

        # Set the feature for unused IOBs, loop over all banks which were
        # observed to have the STEPDOWN feature set.
        missing_features = []

        for bank in stepdown_banks:
            for tile in bank_to_tile[bank]:

                # This is an IOB33 tile. Set the STEPDOWN feature in it but
                # only if it is unused.
                if "IOB33" in tile:
                    for site in get_iob_sites(db, tile):

                        if (tile, site) in used_iob_sites:
                            continue

                        for tag in stepdown_tags[bank]:
                            feature = "{}.{}.{}".format(tile, site, tag)
                            for line in fasm.parse_fasm_string(feature):
                                assembler.add_fasm_line(line, missing_features)

                # This is a HCLK_IOI3 tile, set the stepdown feature for it
                # too.
                if "HCLK_IOI3" in tile:
                    feature = "{}.STEPDOWN".format(tile)
                    for line in fasm.parse_fasm_string(feature):
                        assembler.add_fasm_line(line, missing_features)

        if missing_features:
            raise fasm_assembler.FasmLookupError('\n'.join(missing_features))

    frames = assembler.get_frames(sparse=sparse)

    if debug:
        dump_frames_sparse(frames)

    dump_frm(f_out, frames)


def main():
    parser = argparse.ArgumentParser(
        description=
        'Convert FPGA configuration description ("FPGA assembly") into binary frame equivalent'
    )

    util.db_root_arg(parser)
    util.part_arg(parser)
    parser.add_argument(
        '--sparse', action='store_true', help="Don't zero fill all frames")
    parser.add_argument(
        '--roi',
        help="ROI design.json file defining which tiles are within the ROI.")
    parser.add_argument(
        '--emit_pudc_b_pullup',
        help="Emit an IBUF and PULLUP on the PUDC_B pin if unused",
        action='store_true')
    parser.add_argument(
        '--debug', action='store_true', help="Print debug dump")
    parser.add_argument('fn_in', help='Input FPGA assembly (.fasm) file')
    parser.add_argument(
        'fn_out',
        default='/dev/stdout',
        nargs='?',
        help='Output FPGA frame (.frm) file')

    args = parser.parse_args()
    run(
        db_root=args.db_root,
        part=args.part,
        filename_in=args.fn_in,
        f_out=open(args.fn_out, 'w'),
        sparse=args.sparse,
        roi=args.roi,
        debug=args.debug,
        emit_pudc_b_pullup=args.emit_pudc_b_pullup)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
"""
This script performs the analysis of disassembled bitstream and design information.
It correlates presence/absence of particular fasm features in the bitstream
with presence/absence of a particular IOB type. It also checks for features
that are set for unused IOBs that are in the same bank as the used ones.

The list of available fasm features is read from the prjxray database. The
analysis result is written to a CSV file.
"""
import os
import argparse
import json
import random

from collections import defaultdict

from prjxray import util
import fasm

# =============================================================================


def load_iob_segbits():
    """
    Loads IOB segbits from the database to get fasm feature list.
    This function assumes:
    - LIOB33/RIOB33 have the same features.
    - IOB_Y0 and IOB_Y1 have the same features (at least those of interest).
    """
    features = []

    fname = os.path.join(util.get_db_root(), "segbits_liob33.db")
    with open(fname, "r") as fp:
        for line in fp:

            line = line.strip()
            if len(line) == 0:
                continue

            # Parse the line
            fields = line.split(maxsplit=1)
            feature = fields[0]
            bits = fields[1]

            # Parse the feature
            parts = feature.split(".", maxsplit=3)
            if len(parts) >= 3 and parts[1] == "IOB_Y0":
                features.append(".".join(parts[2:]))

    return features


def correlate_features(features, tile, site, set_features):
    """
    Correlate each feature with the fasm disassembly for given tile/site.

    Given a set of all possible fasm features (for an IOB) in the first
    argument, checks whether they are set or cleared for the given tile+site in
    a design. The parameter set_features contains fasm dissassembly of the
    design.

    Returns a list of tuples with feature names and whether they are set or not.
    """
    result = []

    for feature in features:
        full_feature = "{}.{}.{}".format(tile, site, feature)
        if full_feature in set_features:
            result.append((
                feature,
                True,
            ))
        else:
            result.append((
                feature,
                False,
            ))

    return result


def run():
    """
    Main.
    """

    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--design", type=str, required=True, help="Design JSON file")
    parser.add_argument(
        "--fasm", type=str, required=True, help="Decoded fasm file")
    parser.add_argument(
        "-o", type=str, default="results.csv", help="Output CSV file")
    parser.add_argument("-j", type=str, default=None, help="Output JSON file")

    args = parser.parse_args()

    # Load IOB features
    features = load_iob_segbits()

    # Load the design data
    with open(args.design, "r") as fp:
        design = json.load(fp)

    # Load disassembled fasm
    fasm_tuples = fasm.parse_fasm_filename(args.fasm)
    set_features = fasm.fasm_tuple_to_string(fasm_tuples).split("\n")

    # Correlate features for given IOB types
    results = []
    for region in design:
        result = dict(region["iosettings"])

        for l in ["input", "output", "inout", "unused_sites"]:

            # TODO: Check if this is true eg. for all unused sites, not just
            # one random site.
            tile, site = random.choice(region[l]).split(".")
            matches = correlate_features(features, tile, site, set_features)

            result[l] = matches

        results.append(result)

    # Save results
    if args.j:
        with open(args.j, "w") as fp:
            json.dump(results, fp, indent=2, sort_keys=True)

    # Save results to CSV
    with open(args.o, "w") as fp:
        csv_data = defaultdict(lambda: {})

        # Collect data
        for result in results:
            iostandard = result["iostandard"]
            drive = result["drive"]
            slew = result["slew"]

            if drive is None:
                drive = "_FIXED"

            iosettings = "{}.I{}.{}".format(iostandard, drive, slew)

            is_diff = "DIFF" in iostandard

            for feature in sorted(features):
                I = [f[1] for f in result["input"] if f[0] == feature and f[1]]
                O = [
                    f[1] for f in result["output"] if f[0] == feature and f[1]
                ]
                T = [f[1] for f in result["inout"] if f[0] == feature and f[1]]
                U = [
                    f[1]
                    for f in result["unused_sites"]
                    if f[0] == feature and f[1]
                ]

                s = "".join(
                    [
                        "I" if len(I) > 0 else "",
                        "O" if len(O) > 0 else "",
                        "T" if len(T) > 0 else "",
                        "U" if len(U) > 0 else "",
                    ])

                csv_data[iosettings][feature] = s

        # Write header
        line = ["iosettings"] + sorted(features)
        fp.write(",".join(line) + "\n")

        # Write data
        for iosettings in sorted(csv_data.keys()):
            data = csv_data[iosettings]
            line = [iosettings
                    ] + [data[feature] for feature in sorted(features)]

            fp.write(",".join(line) + "\n")


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
from intervaltree import IntervalTree, Interval
from prjxray import bitstream


class SegmentMap(object):
    def __init__(self, grid):
        self.segment_tree = IntervalTree()

        for bits_info in grid.iter_all_frames():
            self.segment_tree.add(
                Interval(
                    begin=bits_info.bits.base_address,
                    end=bits_info.bits.base_address + bits_info.bits.frames,
                    data=bits_info,
                ))

    def segment_info_for_frame(self, frame):
        """ Return all bits info that match frame address. """
        for frame in self.segment_tree[frame]:
            yield frame.data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os, sys, json

from prjxray.util import OpenSafeFile

def main(argv):
    if len(argv) != 3:
        print("Usage example: python3 %s HCLK_R HCLK_SW6E3" % sys.argv[0])
        sys.exit(1)

    with OpenSafeFile("%s/%s/tileconn.json" % (os.getenv("XRAY_DATABASE_DIR"),
                                       os.getenv("XRAY_DATABASE")), "r") as f:
        tileconn = json.load(f)

    outdata = list()
    max_tiletype_len = 1

    for entry in tileconn:
        if entry["tile_types"][0] == sys.argv[1]:
            this_idx, other_idx = 0, 1
            delta_x, delta_y = entry["grid_deltas"]
        elif entry["tile_types"][1] == sys.argv[1]:
            this_idx, other_idx = 1, 0
            delta_x, delta_y = -entry["grid_deltas"][0], -entry["grid_deltas"][
                1]
        else:
            continue

        for wire_pair in entry["wire_pairs"]:
            if wire_pair[this_idx] != sys.argv[2]:
                continue

            outdata.append(
                (
                    delta_x, delta_y, entry["tile_types"][other_idx],
                    wire_pair[other_idx]))
            max_tiletype_len = max(
                max_tiletype_len, len(entry["tile_types"][other_idx]))

    for entry in outdata:
        print(
            "%3d %3d  %-*s  %s" %
            (entry[0], entry[1], max_tiletype_len, entry[2], entry[3]))


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Generate a primitive to place at every I/O
Unlike CLB tests, the LFSR for this is inside the ROI, not driving it
'''

import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database
import re


def gen_sites():
    '''
    IOB18S: main IOB of a diff pair
    IOB18M: secondary IOB of a diff pair
    IOB18: not a diff pair. Relatively rare (at least in ROI...2 of them?)
    Focus on IOB18S to start
    '''
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        sites = []
        for site_name, site_type in gridinfo.sites.items():
            if site_type == 'IDELAYE2_FINEDELAY':
                sites.append(site_name)

        if len(sites) == 0:
            continue

        sites_y = [
            int(re.match('IDELAY_X[0-9]+Y([0-9]+)', site).group(1))
            for site in sites
        ]

        sites, _ = zip(*sorted(zip(sites, sites_y), key=lambda x: x[1]))

        if gridinfo.tile_type[0] == 'L':
            int_grid_x = loc.grid_x + 3
            pad_grid_x = loc.grid_x - 1
            int_tile_type = 'INT_L'
        else:
            int_grid_x = loc.grid_x - 3
            pad_grid_x = loc.grid_x + 1
            int_tile_type = 'INT_R'

        int_tile_locs = [
            (int_grid_x, loc.grid_y),
        ]

        pad_gridinfo = grid.gridinfo_at_loc((pad_grid_x, loc.grid_y))

        pad_sites = pad_gridinfo.sites.keys()
        pad_sites_y = [
            int(re.match('IOB_X[0-9]+Y([0-9]+)', site).group(1))
            for site in pad_sites
        ]
        pad_sites, _ = zip(
            *sorted(zip(pad_sites, pad_sites_y), key=lambda x: x[1]))

        if not gridinfo.tile_type.endswith("_SING"):
            int_tile_locs.append((int_grid_x, loc.grid_y - 1))

        assert len(sites) == len(int_tile_locs), (
            tile_name, sites, int_tile_locs)
        assert len(sites) == len(pad_sites), (sites, pad_sites)

        for site_name, pad_site, int_tile_loc in zip(sites, pad_sites,
                                                     int_tile_locs):
            int_tile_name = grid.tilename_at_loc(int_tile_loc)
            assert int_tile_name.startswith(int_tile_type), (
                int_tile_name, site_name, int_tile_loc)
            yield int_tile_name, site_name, pad_site


def write_params(params):
    pinstr = ''
    for tile, (site, val, pad_site_name, pin) in sorted(params.items()):
        pinstr += '%s,%s,%s,%s,%s\n' % (tile, val, site, pad_site_name, pin)
    open('params.csv', 'w').write(pinstr)


def run():
    sites = list(gen_sites())
    print(
        '''
`define N_DI {}

module top(input wire [`N_DI-1:0] di);
    wire [`N_DI-1:0] di_buf;

    (* KEEP, DONT_TOUCH, IODELAY_GROUP = "iodelays" *)
    IDELAYCTRL idelayctrl (
        .REFCLK()
    );
    '''.format(len(sites)))

    params = {}

    for idx, ((tile_name, site_name, pad_site_name), isone) in enumerate(zip(
            sites, util.gen_fuzz_states(len(sites)))):
        params[tile_name] = (site_name, isone, pad_site_name, "di[%u]" % idx)

        # Force HARD0 -> GFAN1 with CNTVALUEIN4 = 0
        # Toggle 1 pip with CNTVALUEIN3 = ?
        print(
            '''

            // Solving for {3}
            (* KEEP, DONT_TOUCH *)
            IBUF ibuf_{0}(.I(di[{2}]), .O(di_buf[{2}]));

            (* KEEP, DONT_TOUCH, LOC = "{0}", IODELAY_GROUP = "iodelays" *)
            IDELAYE2 idelay_{0} (
                .CNTVALUEIN(5'b0{1}111),
                .IDATAIN(di_buf[{2}])
                );
'''.format(site_name, isone, idx, tile_name))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2021  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import argparse
import yaml
import subprocess
import os
import re
from prjxray import util


def main():
    """Tool to update all supported, available parts in a mapping file for the
    given family. It will read all parts from Vivado, filter them by the family,
    and will only add these where a device exists for.

    Example:
        prjxray$ ./utils/update_parts.py artix7 --db-root database/artix7/
    """
    parser = argparse.ArgumentParser(
        description="Saves all supported parts for a family.")

    parser.add_argument(
        'family',
        help="Name of the device family.",
        choices=['artix7', 'kintex7', 'zynq7', 'spartan7'])
    util.db_root_arg(parser)

    args = parser.parse_args()
    env = os.environ.copy()
    # Vivado does not use the suffix 7 for zynq
    env['FILTER'] = "zynq" if args.family == "zynq7" else args.family
    cwd = os.path.dirname(os.path.abspath(__file__))
    information = {}

    # Read all supported devices
    supported_devices = util.get_devices(args.db_root).keys()

    # Fetch all parts for a family (FILTER = family)
    command = "{} -mode batch -source update_parts.tcl".format(
        env['XRAY_VIVADO'])
    result = subprocess.run(
        command.split(' '),
        check=True,
        env=env,
        cwd=cwd,
        stdout=subprocess.PIPE)
    parts = result.stdout.decode('utf-8').split('# }\n')[1].splitlines()[:-1]

    # Splits up the part number and checks if the device is supported
    for part in parts:
        part, device, package, speed = part.split(',')
        if device in supported_devices:
            information[part] = {
                'device': device,
                'package': package,
                'speedgrade': speed[1:]
            }
        else:
            print("Part {} has an unsupported device {}".format(part, device))

    # Overwrites the <family>/parts.yaml file completly with new data
    util.set_part_information(args.db_root, information)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
This script generates a verilog ROM module that contains data to be transmitted
and received. The data is random.
'''
import random


def main():

    template = """
`default_nettype none

// ============================================================================

module rom
(
input  wire CLK,
input  wire RST,

input  wire RD,
output wire [{rom_width_minus_one}:0] O
);

// ============================================================================

reg [{rom_width_minus_one}:0] rom[0:{rom_size_minus_one}];

initial begin
{rom_data}
end

reg [{rom_width_minus_one}:0] dat;
reg [{rom_size_bits_minus_one}:0] adr;

always @(posedge CLK)
  if (RST)     adr <= 0;
  else if (RD) adr <= adr + 1;

always @(posedge CLK)
  if (RD) dat <= rom[adr];

assign O = dat;

// ============================================================================

endmodule
"""

    rom_size_bits = 5
    rom_size = 2**rom_size_bits
    rom_width = 8

    rom_data = [random.randint(0, 2**rom_width - 1) for i in range(rom_size)]
    rom_data = "\n".join(
        [
            "  rom[%4d] <= %d'd%d;" % (i, rom_width, d)
            for i, d in enumerate(rom_data)
        ])

    print(
        template.format(
            rom_size_bits_minus_one=rom_size_bits - 1,
            rom_size_minus_one=rom_size - 1,
            rom_width_minus_one=rom_width - 1,
            rom_data=rom_data))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Generate grid from database dump """

from __future__ import print_function
import argparse
import pyjson5 as json5
import multiprocessing
import progressbar
import os.path
import json
import datetime
import pickle
import sys

from prjxray import util, lib
from prjxray.xjson import extract_numbers


def get_tile_grid_info(fname):
    with open(fname, 'r') as f:
        tile = json5.load(f)

    tile_type = tile['type']
    ignored = int(tile['ignored']) != 0

    if ignored:
        tile_type = 'NULL'

    return {
        tile['tile']: {
            'type': tile_type,
            'ignored': ignored,
            'grid_x': tile['x'],
            'grid_y': tile['y'],
            'sites': dict(
                (site['site'], site['type']) for site in tile['sites']),
            'wires': set((wire['wire'] for wire in tile['wires']))
        },
    }


def read_json5(fname):
    with open(fname, 'r') as f:
        return json5.load(f)


def is_edge_shared(edge1, edge2):
    """ Returns true if edge1 or edge2 overlap

  >>> is_edge_shared((0, 1), (0, 1))
  True
  >>> is_edge_shared((0, 2), (0, 1))
  True
  >>> is_edge_shared((0, 1), (0, 2))
  True
  >>> is_edge_shared((1, 2), (0, 3))
  True
  >>> is_edge_shared((0, 3), (1, 2))
  True
  >>> is_edge_shared((1, 2), (0, 2))
  True
  >>> is_edge_shared((0, 2), (1, 2))
  True
  >>> is_edge_shared((0, 2), (1, 3))
  True
  >>> is_edge_shared((1, 3), (0, 2))
  True
  >>> is_edge_shared((0, 1), (1, 2))
  False
  >>> is_edge_shared((1, 2), (0, 1))
  False
  >>> is_edge_shared((0, 1), (2, 3))
  False
  >>> is_edge_shared((2, 3), (0, 1))
  False
  """
    assert edge1[0] < edge1[1], edge1
    assert edge2[0] < edge2[1], edge2

    if edge1[0] <= edge2[0]:
        return edge2[0] < edge1[1]
    else:
        return edge1[0] < edge2[1]


def share_edge(a, b):
    """ Returns true if box defined by a and b share any edge.

  Box is defined as (x-min, y-min, x-max, y-max).

  >>> share_edge((0, 0, 1, 1), (1, 0, 2, 1))
  True
  >>> share_edge((1, 0, 2, 1), (0, 0, 1, 1))
  True
  >>> share_edge((0, 0, 1, 1), (0, 1, 1, 2))
  True
  >>> share_edge((0, 1, 1, 2), (0, 0, 1, 1))
  True
  >>> share_edge((0, 0, 1, 3), (1, 0, 2, 1))
  True
  >>> share_edge((1, 0, 2, 1), (0, 0, 1, 3))
  True
  >>> share_edge((0, 0, 3, 1), (0, 1, 1, 2))
  True
  >>> share_edge((0, 1, 1, 2), (0, 0, 3, 1))
  True
  >>> share_edge((0, 0, 1, 1), (1, 1, 2, 2))
  False
  >>> share_edge((1, 1, 2, 2), (0, 0, 1, 1))
  False
  >>> share_edge((0, 0, 1, 3), (1, 3, 2, 4))
  False
  >>> share_edge((0, 0, 1, 3), (1, 2, 2, 4))
  True
  """

    a_x_min, a_y_min, a_x_max, a_y_max = a
    b_x_min, b_y_min, b_x_max, b_y_max = b

    if a_x_min == b_x_max or a_x_max == b_x_min:
        return is_edge_shared((a_y_min, a_y_max), (b_y_min, b_y_max))
    if a_y_min == b_y_max or a_y_max == b_y_min:
        return is_edge_shared((a_x_min, a_x_max), (b_x_min, b_x_max))


def next_wire_in_dimension(
        wire1, tile1, wire2, tile2, tiles, x_wires, y_wires, wire_map,
        wires_in_node):
    """ next_wire_in_dimension returns true if tile1 and tile2 are in the same
  row and column, and must be adjcent.
  """
    tile1_info = tiles[tile1]
    tile2_info = tiles[tile2]

    tile1_x = tile1_info['grid_x']
    tile2_x = tile2_info['grid_x']
    tile1_y = tile1_info['grid_y']
    tile2_y = tile2_info['grid_y']

    # All wires are in the same row or column or if the each wire lies in its own
    # row or column.
    if len(y_wires) == 1 or len(x_wires) == len(wires_in_node) or abs(
            tile1_y - tile2_y) == 0:
        ordered_wires = sorted(x_wires.keys())

        idx1 = ordered_wires.index(tile1_x)
        idx2 = ordered_wires.index(tile2_x)

        if len(x_wires[tile1_x]) == 1 and len(x_wires[tile2_x]) == 1:
            return abs(idx1 - idx2) == 1

    if len(x_wires) == 1 or len(y_wires) == len(wires_in_node) or abs(
            tile1_x - tile2_x) == 0:
        ordered_wires = sorted(y_wires.keys())

        idx1 = ordered_wires.index(tile1_y)
        idx2 = ordered_wires.index(tile2_y)

        if len(y_wires[tile1_y]) == 1 and len(y_wires[tile2_y]) == 1:
            return abs(idx1 - idx2) == 1

    return None


def only_wire(tile1, tile2, tiles, x_wires, y_wires):
    """ only_wire returns true if tile1 and tile2 only have 1 wire in their respective x or y dimension.
  """
    tile1_info = tiles[tile1]
    tile2_info = tiles[tile2]

    tile1_x = tile1_info['grid_x']
    tile2_x = tile2_info['grid_x']

    tiles_x_adjacent = abs(tile1_x - tile2_x) == 1
    if tiles_x_adjacent and len(x_wires[tile1_x]) == 1 and len(
            x_wires[tile2_x]) == 1:
        return True

    tile1_y = tile1_info['grid_y']
    tile2_y = tile2_info['grid_y']

    tiles_y_adjacent = abs(tile1_y - tile2_y) == 1
    if tiles_y_adjacent and len(y_wires[tile1_y]) == 1 and len(
            y_wires[tile2_y]) == 1:
        return True

    return None


def is_directly_connected(node, node_tree, wire1, wire2):
    if 'wires' in node_tree:
        node_tree_wires = node_tree['wires']
    else:
        if len(node_tree['edges']) == 1 and len(node_tree['joins']) == 0:
            node_tree_wires = node_tree['edges'][0]
        else:
            return None

    if wire1 not in node_tree_wires:
        return None
    if wire2 not in node_tree_wires:
        return None

    # Is there than edge that has wire1 next to wire2?
    for edge in node_tree['edges']:
        idx1 = None
        idx2 = None
        try:
            idx1 = edge.index(wire1)
        except ValueError:
            pass

        try:
            idx2 = edge.index(wire2)
        except ValueError:
            pass

        if idx1 is not None and idx2 is not None:
            return abs(idx1 - idx2) == 1

        if idx1 is not None and (idx1 != 0 and idx1 != len(edge) - 1):
            return False

        if idx2 is not None and (idx2 != 0 and idx2 != len(edge) - 1):
            return False

    # Is there a join of nodes between wire1 and wire2?
    if wire1 in node_tree['joins']:
        return wire2 in node_tree['joins'][wire1]

    if wire2 in node_tree['joins']:
        assert wire1 not in node_tree['joins'][wire2]

    return None


def is_connected(
        wire1, tile1, wire2, tile2, node, wires_in_tiles, wire_map, node_tree,
        tiles, x_wires, y_wires, wires_in_node):
    """ Check if two wires are directly connected. """

    next_wire_in_dim = next_wire_in_dimension(
        wire1, tile1, wire2, tile2, tiles, x_wires, y_wires, wire_map,
        wires_in_node)
    if next_wire_in_dim is not None:
        return next_wire_in_dim

    # Because there are multiple possible wire connections between these two
    # tiles, consult the node_tree to determine if the two wires are actually connected.
    #
    # Warning: The node_tree is incomplete because it is not know how to extract
    # ordered wire information from the node.
    #
    # Example node CLK_BUFG_REBUF_X60Y142/CLK_BUFG_REBUF_R_CK_GCLK0_BOT
    # It does not appear to be possible to get ordered wire connection information
    # for the first two wires connected to PIP
    # CLK_BUFG_REBUF_X60Y117/CLK_BUFG_REBUF.CLK_BUFG_REBUF_R_CK_GCLK0_BOT<<->>CLK_BUFG_REBUF_R_CK_GCLK0_TOP
    #
    # However, it happens to be that theses wires are the only wires in their
    # tiles, so the earlier "only wires in tile" check will pass.

    connected = is_directly_connected(
        node['node'], node_tree[node['node']], wire1, wire2)
    if connected is not None:
        return connected

    is_only_wire = only_wire(tile1, tile2, tiles, x_wires, y_wires)
    if is_only_wire is not None:
        return is_only_wire

    # The node_tree didn't specify these wires, and the wires are not
    # unambiguously connected.
    return False


def process_node(tileconn, key_history, node, wire_map, node_tree, tiles):
    wires = [wire['wire'] for wire in node['wires']]

    wires_in_tiles = {}
    x_wires = {}
    y_wires = {}
    for wire in wires:
        wire_info = wire_map[wire]

        if wire_info['tile'] not in wires_in_tiles:
            wires_in_tiles[wire_info['tile']] = []
        wires_in_tiles[wire_info['tile']].append(wire)

        grid_x = tiles[wire_info['tile']]['grid_x']
        if grid_x not in x_wires:
            x_wires[grid_x] = []
        x_wires[grid_x].append(wire)

        grid_y = tiles[wire_info['tile']]['grid_y']
        if grid_y not in y_wires:
            y_wires[grid_y] = []
        y_wires[grid_y].append(wire)

    if len(wires) == 2:
        wire1 = wires[0]
        wire_info1 = wire_map[wire1]
        wire2 = wires[1]
        wire_info2 = wire_map[wire2]
        update_tile_conn(
            tileconn, key_history, wire1, wire_info1, wire2, wire_info2, tiles)
        return

    for idx, wire1 in enumerate(wires):
        wire_info1 = wire_map[wire1]
        for wire2 in wires[idx + 1:]:
            wire_info2 = wire_map[wire2]

            if not is_connected(wire1, wire_info1['tile'], wire2,
                                wire_info2['tile'], node, wires_in_tiles,
                                wire_map, node_tree, tiles, x_wires, y_wires,
                                wires):
                continue

            update_tile_conn(
                tileconn, key_history, wire1, wire_info1, wire2, wire_info2,
                tiles)


def update_tile_conn(
        tileconn, key_history, wirename1, wire1, wirename2, wire2, tiles):
    # Ensure that (wire1, wire2) is sorted, so we can easy check if a connection
    # already exists.

    tile1 = tiles[wire1['tile']]
    tile2 = tiles[wire2['tile']]
    if ((wire1['type'], wire1['shortname'], tile1['grid_x'], tile1['grid_y']) >
        (wire2['type'], wire2['shortname'], tile2['grid_x'], tile2['grid_y'])):
        wire1, tile1, wire2, tile2 = wire2, tile2, wire1, tile1

    tileconn.append(
        {
            "grid_deltas": [
                tile2['grid_x'] - tile1['grid_x'],
                tile2['grid_y'] - tile1['grid_y'],
            ],
            "tile_types": [
                tile1['type'],
                tile2['type'],
            ],
            "wire_pair": [
                wire1['shortname'],
                wire2['shortname'],
            ],
        })


def flatten_tile_conn(tileconn):
    """ Convert tileconn that is key'd to identify specific wire pairs between tiles
  key (tile1_type, wire1_name, tile2_type, wire2_name) to flat tile connect list
  that relates tile types and relative coordinates and a full list of wires to
  connect. """
    flat_tileconn = {}

    for conn in tileconn:
        key = (tuple(conn['tile_types']), tuple(conn['grid_deltas']))

        if key not in flat_tileconn:
            flat_tileconn[key] = {
                'tile_types': conn['tile_types'],
                'grid_deltas': conn['grid_deltas'],
                'wire_pairs': set()
            }

        flat_tileconn[key]['wire_pairs'].add(tuple(conn['wire_pair']))

    def inner():
        for output in flat_tileconn.values():
            yield {
                'tile_types': output['tile_types'],
                'grid_deltas': output['grid_deltas'],
                'wire_pairs': tuple(output['wire_pairs']),
            }

    return tuple(inner())


def is_tile_type(tiles, coord_to_tile, coord, tile_type):
    if coord not in coord_to_tile:
        return False

    target_tile = tiles[coord_to_tile[coord]]
    return target_tile['type'] == tile_type


def get_connections(wire, wire_info, conn, idx, coord_to_tile, tiles):
    """ Yields (tile_coord, wire) for each wire that should be connected to specified wire. """
    pair = conn['wire_pairs'][idx]
    wire_tile_type = wire_info['type']
    tile_types = conn['tile_types']
    shortname = wire_info['shortname']
    grid_deltas = conn['grid_deltas']

    wire1 = tile_types[0] == wire_tile_type and shortname == pair[0]
    wire2 = tile_types[1] == wire_tile_type and shortname == pair[1]
    assert wire1 or wire2, (wire, conn)

    tile_of_wire = wire_info['tile']
    start_coord_x = tiles[tile_of_wire]['grid_x']
    start_coord_y = tiles[tile_of_wire]['grid_y']
    if wire1:
        target_coord_x = start_coord_x + grid_deltas[0]
        target_coord_y = start_coord_y + grid_deltas[1]
        target_tile_type = tile_types[1]

        target_wire = pair[1]
        target_tile = (target_coord_x, target_coord_y)

        if is_tile_type(tiles, coord_to_tile, target_tile, target_tile_type):
            yield target_tile, target_wire

    if wire2:
        target_coord_x = start_coord_x - grid_deltas[0]
        target_coord_y = start_coord_y - grid_deltas[1]
        target_tile_type = tile_types[0]

        target_wire = pair[0]
        target_tile = (target_coord_x, target_coord_y)

        if is_tile_type(tiles, coord_to_tile, target_tile, target_tile_type):
            yield target_tile, target_wire


def make_connection(wire_nodes, wire1, wire2):
    if wire_nodes[wire1] is wire_nodes[wire2]:
        assert wire1 in wire_nodes[wire1]
        assert wire2 in wire_nodes[wire2]
        return

    new_node = wire_nodes[wire1] | wire_nodes[wire2]

    for wire in new_node:
        wire_nodes[wire] = new_node


def create_coord_to_tile(tiles):
    coord_to_tile = {}
    for tile, tileinfo in tiles.items():
        coord_to_tile[(tileinfo['grid_x'], tileinfo['grid_y'])] = tile

    return coord_to_tile


def connect_wires(tiles, tileconn, wire_map):
    """ Connect individual wires into groups of wires called nodes. """

    # Initialize all nodes to originally only contain the wire by itself.
    wire_nodes = {}
    for wire in wire_map:
        wire_nodes[wire] = set([wire])

    wire_connection_map = {}
    for conn in tileconn:
        for idx, (wire1, wire2) in enumerate(conn['wire_pairs']):
            key1 = (conn['tile_types'][0], wire1)
            if key1 not in wire_connection_map:
                wire_connection_map[key1] = []
            wire_connection_map[key1].append((conn, idx))

            key2 = (conn['tile_types'][1], wire2)
            if key2 not in wire_connection_map:
                wire_connection_map[key2] = []
            wire_connection_map[key2].append((conn, idx))

    coord_to_tile = create_coord_to_tile(tiles)

    for wire, wire_info in progressbar.progressbar(wire_map.items()):
        key = (wire_info['type'], wire_info['shortname'])
        if key not in wire_connection_map:
            continue

        for conn, idx in wire_connection_map[key]:
            for target_tile, target_wire in get_connections(
                    wire, wire_info, conn, idx, coord_to_tile, tiles):

                full_wire_name = coord_to_tile[target_tile] + '/' + target_wire
                assert wire_map[full_wire_name]['shortname'] == target_wire, (
                    target_tile, target_wire, wire, conn)
                assert wire_map[full_wire_name]['tile'] == coord_to_tile[
                    target_tile], (
                        wire_map[full_wire_name]['tile'],
                        coord_to_tile[target_tile])

                make_connection(wire_nodes, wire, full_wire_name)

    # Find unique nodes
    nodes = {}
    for node in wire_nodes.values():
        nodes[id(node)] = node

    # Flatten to list of lists.
    return tuple(tuple(node) for node in nodes.values())


def generate_tilegrid(pool, tiles):
    wire_map = {}

    grid = {}

    num_tiles = 0
    for tile_type in tiles:
        num_tiles += len(tiles[tile_type])

    idx = 0
    with progressbar.ProgressBar(max_value=num_tiles) as bar:
        for tile_type in tiles:
            for tile in pool.imap_unordered(
                    get_tile_grid_info,
                    tiles[tile_type],
                    chunksize=20,
            ):
                bar.update(idx)

                assert len(tile) == 1, tile
                tilename = tuple(tile.keys())[0]

                for wire in tile[tilename]['wires']:
                    assert wire not in wire_map, (wire, wire_map)
                    assert wire.startswith(tilename + '/'), (wire, tilename)

                    wire_map[wire] = {
                        'tile': tilename,
                        'type': tile[tilename]['type'],
                        'shortname': wire[len(tilename) + 1:],
                    }

                del tile[tilename]['wires']
                grid.update(tile)

                idx += 1
                bar.update(idx)

    return grid, wire_map


def generate_tileconn(pool, node_tree, nodes, wire_map, grid):
    tileconn = []
    key_history = {}
    raw_node_data = []
    with progressbar.ProgressBar(max_value=len(nodes)) as bar:
        for idx, node in enumerate(pool.imap_unordered(
                read_json5,
                nodes,
                chunksize=20,
        )):
            bar.update(idx)
            raw_node_data.append(node)
            process_node(
                tileconn, key_history, node, wire_map, node_tree, grid)
            bar.update(idx + 1)

    tileconn = flatten_tile_conn(tileconn)

    return tileconn, raw_node_data


def main():
    parser = argparse.ArgumentParser(
        description=
        "Reduces raw database dump into prototype tiles, grid, and connections."
    )
    parser.add_argument('--root_dir', required=True)
    parser.add_argument('--output_dir', required=True)
    parser.add_argument('--verify_only', action='store_true')
    parser.add_argument('--ignored_wires')
    parser.add_argument('--max_cpu', type=int, default=10)

    args = parser.parse_args()

    tiles, nodes = lib.read_root_csv(args.root_dir)

    processes = min(multiprocessing.cpu_count(), args.max_cpu)
    print('{} Running {} processes'.format(datetime.datetime.now(), processes))
    pool = multiprocessing.Pool(processes=processes)

    node_tree_file = os.path.join(args.output_dir, 'node_tree.json')

    tileconn_file = os.path.join(args.output_dir, 'tileconn.json')
    wire_map_file = os.path.join(args.output_dir, 'wiremap.pickle')

    print('{} Reading tilegrid'.format(datetime.datetime.now()))
    with open(os.path.join(util.get_db_root(), util.get_fabric(),
                           'tilegrid.json')) as f:
        grid = json.load(f)

    if not args.verify_only:
        print('{} Creating tile map'.format(datetime.datetime.now()))
        grid2, wire_map = generate_tilegrid(pool, tiles)

        # Make sure tilegrid from 005-tilegrid matches tilegrid from
        # generate_tilegrid.
        db_grid_keys = set(grid.keys())
        generated_grid_keys = set(grid2.keys())
        assert db_grid_keys == generated_grid_keys, (
            db_grid_keys ^ generated_grid_keys)

        for tile in db_grid_keys:
            for k in grid2[tile]:
                if k == 'ignored':
                    continue

                if k == 'sites' and grid2[tile]['ignored']:
                    continue

                assert k in grid[tile], k
                assert grid[tile][k] == grid2[tile][k], (
                    tile, k, grid[tile][k], grid2[tile][k])

        with open(wire_map_file, 'wb') as f:
            pickle.dump(wire_map, f)

        print('{} Reading node tree'.format(datetime.datetime.now()))
        with open(node_tree_file) as f:
            node_tree = json.load(f)

        print('{} Creating tile connections'.format(datetime.datetime.now()))
        tileconn, raw_node_data = generate_tileconn(
            pool, node_tree, nodes, wire_map, grid)

        for data in tileconn:
            data['wire_pairs'] = tuple(
                sorted(
                    data['wire_pairs'],
                    key=lambda x: tuple(extract_numbers(s) for s in x)))

        tileconn = tuple(
            sorted(
                tileconn, key=lambda x: (x['tile_types'], x['grid_deltas'])))

        print('{} Writing tileconn'.format(datetime.datetime.now()))
        with open(tileconn_file, 'w') as f:
            json.dump(tileconn, f, indent=2, sort_keys=True)
    else:
        with open(wire_map_file, 'rb') as f:
            wire_map = pickle.load(f)

        print('{} Reading raw_node_data'.format(datetime.datetime.now()))
        raw_node_data = []
        with progressbar.ProgressBar(max_value=len(nodes)) as bar:
            for idx, node in enumerate(pool.imap_unordered(
                    read_json5,
                    nodes,
                    chunksize=20,
            )):
                bar.update(idx)
                raw_node_data.append(node)
                bar.update(idx + 1)

        print('{} Reading tileconn'.format(datetime.datetime.now()))
        with open(tileconn_file) as f:
            tileconn = json.load(f)

    wire_nodes_file = os.path.join(args.output_dir, 'wire_nodes.pickle')
    if os.path.exists(wire_nodes_file) and args.verify_only:
        with open(wire_nodes_file, 'rb') as f:
            wire_nodes = pickle.load(f)
    else:
        print(
            "{} Connecting wires to verify tileconn".format(
                datetime.datetime.now()))
        wire_nodes = connect_wires(grid, tileconn, wire_map)
        with open(wire_nodes_file, 'wb') as f:
            pickle.dump(wire_nodes, f)

    print('{} Verifing tileconn'.format(datetime.datetime.now()))
    error_nodes = []
    lib.verify_nodes(
        [
            (node['node'], tuple(wire['wire']
                                 for wire in node['wires']))
            for node in raw_node_data
        ], wire_nodes, error_nodes)

    if len(error_nodes) > 0:
        error_nodes_file = os.path.join(args.output_dir, 'error_nodes.json')
        with open(error_nodes_file, 'w') as f:
            json.dump(error_nodes, f, indent=2, sort_keys=True)

        ignored_wires = []
        ignored_wires_file = args.ignored_wires
        if os.path.exists(ignored_wires_file):
            with open(ignored_wires_file) as f:
                ignored_wires = set(l.strip() for l in f)

        if not lib.check_errors(error_nodes, ignored_wires):
            print(
                '{} errors detected, see {} for details.'.format(
                    len(error_nodes), error_nodes_file))
            sys.exit(1)
        else:
            print(
                '{} errors ignored because of {}\nSee {} for details.'.format(
                    len(error_nodes), ignored_wires_file, error_nodes_file))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray.db import Database
from prjxray import util
from prjxray.lut_maker import LutMaker


def gen_sites():
    xy_fun = util.create_xy_fun('BUFR_')
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        sites = []

        xs = []
        ys = []
        for site, site_type in gridinfo.sites.items():
            if site_type == 'BUFR':
                x, y = xy_fun(site)
                xs.append(x)
                ys.append(y)

                sites.append((site, x, y))

        if not sites:
            continue

        ioi3 = grid.gridinfo_at_loc((loc.grid_x, loc.grid_y - 1))

        if 'IOI3' not in ioi3.tile_type:
            continue

        if ioi3.tile_type.startswith('R'):
            dx = 1
        else:
            assert ioi3.tile_type.startswith('L')
            dx = -1

        iobs = []

        for dy in (-1, -3, 2, 4):
            iob = grid.gridinfo_at_loc((loc.grid_x + dx, loc.grid_y + dy))

            for site, site_type in iob.sites.items():
                if site_type == 'IOB33M':
                    iobs.append(site)

        yield tile_name, min(xs), min(ys), sorted(sites), sorted(iobs)


def main():

    params_list = []
    num_clocks = 0
    outputs = []
    luts = LutMaker()
    for tile_name, x_min, y_min, sites, iobs in gen_sites():
        ioclks = []
        for iob in iobs:
            ioclk = 'clk_{}'.format(iob)
            ioclks.append(ioclk)
            idx = num_clocks
            num_clocks += 1
            outputs.append(
                '''
        wire {ioclk};

        (* KEEP, DONT_TOUCH, LOC="{site}" *)
        IBUF #(
            .IOSTANDARD("LVCMOS33")
            ) ibuf_{site} (
                .I(clks[{idx}]),
                .O({ioclk})
                );'''.format(
                    ioclk=ioclk,
                    site=iob,
                    idx=idx,
                ))

        for site, x, y in sites:
            params = {}
            params['tile'] = tile_name
            params['site'] = site
            params['IN_USE'] = random.randint(0, 1)
            params['x'] = x - x_min
            params['y'] = y - y_min

            if params['IN_USE']:
                params['BUFR_DIVIDE'] = random.choice(
                    (
                        '"BYPASS"',
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                    ))
                params['I'] = random.choice(ioclks)

                if params['BUFR_DIVIDE'] == '"BYPASS"':
                    params['CE'] = '1'
                    params['CLR'] = '0'
                else:
                    params['CE'] = luts.get_next_output_net()
                    params['CLR'] = luts.get_next_output_net()

                outputs.append(
                    '''
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    BUFR #(
        .BUFR_DIVIDE({BUFR_DIVIDE})
        ) buf_{site} (
            .CE({CE}),
            .CLR({CLR}),
            .I({I})
        );
                    '''.format(**params))

            params_list.append(params)

    print(
        '''
module top(input [{n1}:0] clks);
    '''.format(n1=num_clocks - 1))

    print("""
    (* KEEP, DONT_TOUCH *)
    LUT6 dummy (
        );""")

    for l in luts.create_wires_and_luts():
        print(l)

    for l in outputs:
        print(l)

    print("endmodule")

    with open('params.json', 'w') as f:
        json.dump(params_list, f, indent=2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" This script creates node_wires.json, which describes how nodes are named.

This script consumes the raw node data from root_dir and outputs
node_wires.json to the output_dir.

The class prjxray.node_model.NodeModel can be used to reconstruct node names
and node <-> wire mapping.

The contents of node_wires.json is:
 - The set of tile type wires that are always nodes, key "node_pattern_wires"
 - The set of tile wires that are nodes within the graph, key
   "specific_node_wires".

"""

import argparse
import datetime
import json
import multiprocessing
import progressbar
import pyjson5 as json5
import os.path

from prjxray import util, lib
from prjxray.grid import Grid


def read_json5(fname):
    with open(fname, 'r') as f:
        return json5.load(f)


def main():
    parser = argparse.ArgumentParser(
        description="Reduce node names for wire connections.")
    parser.add_argument('--root_dir', required=True)
    parser.add_argument('--output_dir', required=True)
    parser.add_argument('--max_cpu', type=int, default=10)

    args = parser.parse_args()

    _, nodes = lib.read_root_csv(args.root_dir)

    processes = min(multiprocessing.cpu_count(), args.max_cpu)
    pool = multiprocessing.Pool(processes=processes)

    # Read tile grid and raw node data.
    print('{} Reading tilegrid'.format(datetime.datetime.now()))
    with open(os.path.join(util.get_db_root(), util.get_fabric(),
                           'tilegrid.json')) as f:
        grid = Grid(db=None, tilegrid=json.load(f))

    raw_node_data = []
    with progressbar.ProgressBar(max_value=len(nodes)) as bar:
        for idx, node in enumerate(pool.imap_unordered(
                read_json5,
                nodes,
                chunksize=20,
        )):
            bar.update(idx)
            raw_node_data.append(node)
            bar.update(idx + 1)

    node_wires = set()
    remove_node_wires = set()
    specific_node_wires = set()

    # Create initial node wire pattern
    for node in progressbar.progressbar(raw_node_data):
        if len(node['wires']) <= 1:
            continue

        node_tile, node_wire = node['node'].split('/')

        for wire in node['wires']:
            wire_tile, wire_name = wire['wire'].split('/')

            if node['node'] == wire['wire']:
                assert node_tile == wire_tile
                assert node_wire == wire_name
                gridinfo = grid.gridinfo_at_tilename(node_tile)
                node_wires.add((gridinfo.tile_type, wire_name))

    print(
        'Initial number of wires that are node drivers: {}'.format(
            len(node_wires)))

    # Remove exceptional node wire names, create specific_node_wires set,
    # which is simply the list of wires that are nodes in the graph.
    for node in progressbar.progressbar(raw_node_data):
        if len(node['wires']) <= 1:
            continue

        for wire in node['wires']:
            wire_tile, wire_name = wire['wire'].split('/')
            gridinfo = grid.gridinfo_at_tilename(wire_tile)
            key = gridinfo.tile_type, wire_name

            if node['node'] == wire['wire']:
                assert key in node_wires
            else:
                if key in node_wires:
                    specific_node_wires.add(node['node'])
                    remove_node_wires.add(key)

    # Complete the specific_node_wires list after the pruning of the
    # node_pattern_wires sets.
    for node in progressbar.progressbar(raw_node_data):
        if len(node['wires']) <= 1:
            continue

        for wire in node['wires']:
            wire_tile, wire_name = wire['wire'].split('/')
            gridinfo = grid.gridinfo_at_tilename(wire_tile)
            key = gridinfo.tile_type, wire_name

            if key in remove_node_wires and node['node'] == wire['wire']:
                specific_node_wires.add(node['node'])

    node_wires -= remove_node_wires
    print(
        'Final number of wires that are node drivers: {}'.format(
            len(node_wires)))
    print(
        'Number of wires that are node drivers: {}'.format(
            len(specific_node_wires)))

    # Verify the node wire data.
    for node in progressbar.progressbar(raw_node_data):
        if len(node['wires']) <= 1:
            continue

        found_node_wire = False
        for wire in node['wires']:
            if wire['wire'] in specific_node_wires:
                assert wire['wire'] == node['node']

                found_node_wire = True
                break

        if not found_node_wire:
            for wire in node['wires']:
                wire_tile, wire_name = wire['wire'].split('/')
                gridinfo = grid.gridinfo_at_tilename(wire_tile)
                key = gridinfo.tile_type, wire_name

                if key in node_wires:
                    assert node['node'] == wire['wire']
                else:
                    assert node['node'] != wire['wire']

    # Normalize output.
    tile_types = {}
    for tile_type, tile_wire in node_wires:
        if tile_type not in tile_types:
            tile_types[tile_type] = []

        tile_types[tile_type].append(tile_wire)

    for tile_type in tile_types:
        tile_types[tile_type].sort()

    out = {
        'node_pattern_wires': tile_types,
        'specific_node_wires': sorted(specific_node_wires),
    }

    with open(os.path.join(args.output_dir, 'node_wires.json'), 'w') as f:
        json.dump(out, f, indent=2, sort_keys=True)


if __name__ == '__main__':
    main()
